# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#
import logging
import os.path
import sqlite3

from osdlyrics.consts import (METADATA_ALBUM, METADATA_ARTIST, METADATA_TITLE,
                              METADATA_TRACKNUM)
import osdlyrics.utils

__all__ = (
    'LrcDb',
)


def query_param_from_metadata(metadata):
    """
    Generate query dict from metadata
    """
    param = {
        METADATA_TITLE: metadata.title or '',
        METADATA_ARTIST: metadata.artist or '',
        METADATA_ALBUM: metadata.album or '',
        METADATA_TRACKNUM: max(metadata.tracknum, 0),
    }
    return param


class LrcDb:
    """ Database to store location of LRC files that have been manually assigned
    """

    TABLE_NAME = 'lyrics'

    METADATA_LIST = [METADATA_TITLE, METADATA_ARTIST, METADATA_ALBUM, METADATA_TRACKNUM]

    CREATE_TABLE = """
CREATE TABLE IF NOT EXISTS {0} (
  id INTEGER PRIMARY KEY AUTOINCREMENT,
  {1} TEXT, {2} TEXT, {3} TEXT, {4} INTEGER,
  uri TEXT UNIQUE ON CONFLICT REPLACE,
  lrcpath TEXT
)
""".format(TABLE_NAME, *METADATA_LIST)

    ASSIGN_LYRIC = """
INSERT OR REPLACE INTO {0}
  ({1}, {2}, {3}, {4}, uri, lrcpath)
  VALUES (?, ?, ?, ?, ?, ?)
""" .format(TABLE_NAME, *METADATA_LIST)

    UPDATE_LYRIC = """
UPDATE {0}
  SET lrcpath=?
  WHERE uri=?
""".format(TABLE_NAME)

    DELETE_LYRIC = 'DELETE FROM {0} WHERE '.format(TABLE_NAME)

    FIND_LYRIC = 'SELECT lrcpath FROM {0} WHERE '.format(TABLE_NAME)

    QUERY_LOCATION = 'uri = ?'

    QUERY_INFO = ' AND '.join('{0}=:{0}'.format(m) for m in METADATA_LIST)

    def __init__(self, dbfile=None):
        """

        Arguments:
        - `dbfile`: The sqlite db to open
        """
        if dbfile is None:
            dbfile = osdlyrics.utils.get_config_path('lrc.db')
        self._dbfile = dbfile
        osdlyrics.utils.ensure_path(dbfile)
        self._conn = sqlite3.connect(os.path.expanduser(dbfile))
        self._create_table()

    def _create_table(self):
        """ Ensures the table structure of new open dbs
        """
        c = self._conn.cursor()
        c.execute(LrcDb.CREATE_TABLE)
        self._conn.commit()
        c.close()

    def assign(self, metadata, uri):
        # type: (osdlyrics.metadata.Metadata, Text) -> None
        """ Assigns a uri of lyrics to tracks represented by metadata
        """
        c = self._conn.cursor()
        location = metadata.location or ''
        if self._find_by_location(metadata):
            logging.debug('Assign lyric file %s to track of location %s', uri, location)
            c.execute(LrcDb.UPDATE_LYRIC, (uri, location,))
        else:
            title = metadata.title or ''
            artist = metadata.artist or ''
            album = metadata.album or ''
            tracknum = max(metadata.tracknum, 0)
            logging.debug('Assign lyrics file %s to track %s. %s - %s in album %s @ %s', uri, tracknum, artist, title, album, location)
            c.execute(LrcDb.ASSIGN_LYRIC, (title, artist, album, tracknum, location, uri))
        self._conn.commit()
        c.close()

    def delete(self, metadata):
        """ Deletes lyrics association(s) for given metadata

        Deletes all lyrics associations that would be found by find(self, metadata)
        """
        c = self._conn.cursor()

        if metadata.location:
            c.execute(LrcDb.DELETE_LYRIC + LrcDb.QUERY_LOCATION, (metadata.location,))

        c.execute(LrcDb.DELETE_LYRIC + LrcDb.QUERY_INFO, query_param_from_metadata(metadata))

        self._conn.commit()
        c.close()

    def find(self, metadata):
        """ Finds the location of LRC files for given metadata

        To find the location of lyrics, firstly find whether there is a record matched
        with the ``location`` attribute in metadata. If not found or ``location`` is
        not specified, try to find with respect to ``title``, ``artist``, ``album``
        and ``tracknum``

        If found, return the uri of the LRC file. Otherwise return None. Note that
        this method may return an empty string, so use ``is None`` to figure out
        whether an uri is found
        """
        ret = self._find_by_location(metadata)
        if ret is not None:
            return ret
        ret = self._find_by_info(metadata)
        if ret is not None:
            return ret
        return None

    def _find_by_condition(self, where_clause, parameters=None):
        query = LrcDb.FIND_LYRIC + where_clause
        logging.debug('Find by condition, query = %s, params = %s', query, parameters)
        c = self._conn.cursor()
        c.execute(query, parameters)
        r = c.fetchone()
        logging.debug('Fetch result: %s', r)
        if r:
            return r[0]
        return None

    def _find_by_location(self, metadata):
        if not metadata.location:
            return None
        return self._find_by_condition(LrcDb.QUERY_LOCATION, (metadata.location,))

    def _find_by_info(self, metadata):
        return self._find_by_condition(LrcDb.QUERY_INFO,
                                       query_param_from_metadata(metadata))


def test():
    """
    >>> import dbus
    >>> from osdlyrics.metadata import Metadata
    >>> db = LrcDb('/tmp/asdf')
    >>> db.assign(Metadata.from_dict({'title': 'Tiger',
    ...                               'artist': 'Soldier',
    ...                               'location': 'file:///tmp/asdf'}),
    ...           'file:///tmp/a.lrc')
    >>> db.find(Metadata.from_dict({'location': 'file:///tmp/asdf'}))
    'file:///tmp/a.lrc'
    >>> db.find(Metadata.from_dict({'location': 'file:///tmp/asdfg'}))
    >>> db.find(Metadata.from_dict({'title': 'Tiger',
    ...                             'location': 'file:///tmp/asdf'}))
    'file:///tmp/a.lrc'
    >>> db.find(Metadata.from_dict({'title': 'Tiger', }))
    >>> db.find(Metadata.from_dict({'title': 'Tiger',
    ...                             'artist': 'Soldier'}))
    'file:///tmp/a.lrc'
    >>> db.assign(Metadata.from_dict({'title': 'ttTiger',
    ...                               'artist': 'ssSoldier',
    ...                               'location': 'file:///tmp/asdf'}),
    ...           'file:///tmp/b.lrc')
    >>> db.find(Metadata.from_dict({'artist': 'Soldier', }))
    >>> db.find(Metadata.from_dict({'title': 'Tiger',
    ...                                      'artist': 'Soldier', }))
    'file:///tmp/b.lrc'
    >>> db.find(Metadata.from_dict({dbus.String('title'): dbus.String('Tiger'),
    ...                             dbus.String('artist'): dbus.String('Soldier'), }))
    'file:///tmp/b.lrc'
    >>> metadata_uni = Metadata.from_dict({'title': '\u6807\u9898', 'artist': '\u6b4c\u624b', })
    >>> db.assign(metadata_uni, '\u8def\u5f84')
    >>> db.find(metadata_uni)
    '\u8def\u5f84'
    >>> db.find(Metadata.from_dict({'title': 'Tiger', 'artist': 'Soldiers', }))
    >>> db.find(Metadata())
    >>> db.delete(Metadata.from_dict({'location': 'file:///tmp/asdf'}))
    >>> db.find(Metadata.from_dict({'title': 'Tiger',
    ...                             'artist': 'Soldier',
    ...                             'location': 'file:///tmp/asdf'}))
    """
    import doctest
    doctest.testmod()


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

from abc import abstractmethod
import logging

import dbus
import dbus.service

from . import errors, timer
from .app import App
from .consts import (MPRIS2_PLAYER_INTERFACE, PLAYER_PROXY_INTERFACE,
                     PLAYER_PROXY_OBJECT_PATH_PREFIX)
from .dbusext.service import Object as DBusObject, property as dbus_property


class CAPS:
    NEXT = 1 << 0
    PREV = 1 << 1
    PAUSE = 1 << 2
    PLAY = 1 << 3
    SEEK = 1 << 4
    PROVIDE_METADATA = 1 << 5


class REPEAT:
    NONE = 0
    TRACK = 1
    ALL = 2


class STATUS:
    PLAYING = 0
    PAUSED = 1
    STOPPED = 2


class ConnectPlayerError(errors.BaseError):
    """
    Exception raised when BasePlayerProxy.do_connect_player() fails
    """
    pass


class BasePlayerProxy(dbus.service.Object):
    """ Base class to create an application to provide player proxy support
    """

    def __init__(self, name):
        """

        Arguments:
        - `name`: The suffix of the bus name. The full bus name is
          `org.osdlyrics.PlayerProxy.` + name
        """
        self._app = App('PlayerProxy.' + name)
        super().__init__(conn=self._app.connection,
                         object_path=PLAYER_PROXY_OBJECT_PATH_PREFIX + name)
        self._name = name
        self._connected_players = {}

    @property
    def name(self):
        return self._name

    def run(self):
        self._app.run()

    @dbus.service.method(dbus_interface=PLAYER_PROXY_INTERFACE,
                         in_signature='',
                         out_signature='aa{sv}')
    def ListActivePlayers(self):
        return [player.to_dict() for player in self.do_list_active_players()]

    @dbus.service.method(dbus_interface=PLAYER_PROXY_INTERFACE,
                         in_signature='',
                         out_signature='aa{sv}')
    def ListSupportedPlayers(self):
        return [player.to_dict() for player in self.do_list_supported_players()]

    @dbus.service.method(dbus_interface=PLAYER_PROXY_INTERFACE,
                         in_signature='',
                         out_signature='aa{sv}')
    def ListActivatablePlayers(self):
        return [player.to_dict() for player in
                self.do_list_activatable_players()]

    @dbus.service.method(dbus_interface=PLAYER_PROXY_INTERFACE,
                         in_signature='s',
                         out_signature='o')
    def ConnectPlayer(self, player_name):
        if self._connected_players.setdefault(player_name, None):
            return self._connected_players[player_name].object_path
        try:
            player = self.do_connect_player(player_name)
        except TypeError as e:
            raise errors.BaseError(
                '%s cannot instantiate Player[%s, %s]' % (type(self).__name__, self.name, player_name)
            ) from e
        if player and player.connected:
            player.set_disconnect_cb(self._player_lost_cb)
            self._connected_players[player_name] = player
            logging.info('Connected to %s', player.object_path)
            return player.object_path
        else:
            raise ConnectPlayerError('%s cannot be connected', player_name)

    @dbus.service.signal(dbus_interface=PLAYER_PROXY_INTERFACE,
                         signature='s')
    def PlayerLost(self, player_name):
        pass

    def _player_lost_cb(self, player):
        if player.name in self._connected_players:
            del self._connected_players[player.name]
            self.PlayerLost(player.name)

    @abstractmethod
    def do_list_active_players(self):
        """
        Lists supported players that are aready running

        Returns an list of `PlayerInfo` objects.
        """
        raise NotImplementedError()

    @abstractmethod
    def do_list_supported_players(self):
        """
        Lists supported players.

        Returns an list of `PlayerInfo` objects.
        """
        raise NotImplementedError()

    @abstractmethod
    def do_list_activatable_players(self):
        """
        Lists supported players installed on the system.

        Returns an list of `PlayerInfo` objects.
        """
        raise NotImplementedError()

    def do_connect_player(self, playername):
        """
        Creates an Player object according to playername.

        Returns the created `BasePlayer` object, or None if cannot connect to
        the player with `playername`.
        """
        raise NotImplementedError()


class PlayerInfo:
    """Information about a supported player
    """

    def __init__(self, name, appname='', binname='', cmd='', icon=''):
        self._name = name
        self._appname = appname
        self._binname = binname
        self._cmd = cmd
        self._icon = icon

    @classmethod
    def from_name(cls, name):
        return cls(name, icon=name)

    @property
    def name(self):
        return self._name

    @property
    def appname(self):
        return self._appname

    @property
    def binname(self):
        return self._binname

    @property
    def cmd(self):
        return self._cmd

    @property
    def icon(self):
        return self._icon

    def to_dict(self):
        """
        Converts the PlayerInfo object to an dict that fits the specification
        """
        keys = ['name', 'appname', 'binname', 'cmd', 'icon']
        ret = {}
        for k in keys:
            ret[k] = getattr(self, '_' + k)
        return ret


class BasePlayer(DBusObject):
    """ Base class of a player

    Derived classes MUST reimplement following methods:

    - `get_metadata`
    - `get_position`
    - `get_caps`

    Derived classes SHOULD reimplement following methods if supported.
    - `get_status`
    - `get_repeat`
    - `set_repeat`
    - `get_shuffle`
    - `set_shuffle`
    - `play`
    - `pause`
    - `prev`
    - `next`
    - `set_position`
    - `set_volume`
    - `get_volume`
    """

    def __init__(self, proxy, name):
        """

        Arguments:
        - `proxy`: The BasePlayerProxy object that creates the player
        - `name`: The name of the player object
        """
        self._object_path = PLAYER_PROXY_OBJECT_PATH_PREFIX + proxy.name + '/' + name
        super().__init__(conn=proxy.connection,
                         object_path=self._object_path)
        self._name = name
        self._proxy = proxy
        self._disconnect_cb = None
        self._connected = True
        self._timer = None
        self._status = None
        self._loop_status = None
        self._metadata = None
        self._current_trackid = 0
        self._caps = None
        self._shuffle = None

    def set_disconnect_cb(self, disconnect_cb):
        self._disconnect_cb = disconnect_cb

    @property
    def name(self):
        return self._name

    @property
    def proxy(self):
        return self._proxy

    def disconnect(self):
        if self._connected:
            self._connected = False
            self.remove_from_connection()
            if callable(self._disconnect_cb):
                self._disconnect_cb(self)

    def get_status(self):
        """
        Return the playing status.

        The return value should be one of STATUS.PLAYING, STATUS.PAUSED, and
        STATUS.STOPPED

        Derived classes that supports playing status should reimplement this.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_metadata(self):
        # type: () -> Metadata
        """
        Return metadata of current track. The return value is of the type Metadata
        """
        raise NotImplementedError()

    @abstractmethod
    def get_position(self):
        """
        Gets the ellapsed time in current track.

        Returns the time in milliseconds.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_caps(self):
        """
        Return capablities of the players.

        The return value should be a set of CAPS.PLAY, CAPS.PAUSE, CAPS.NEXT,
        CAPS.PREV, CAPS.SEEK
        """
        raise NotImplementedError()

    def get_repeat(self):
        """
        Gets the repeat mode of the player

        Returns one of REPEAT.NONE, REPEAT.TRACK, or REPEAT.ALL

        The default implementation returns REPEAT.NONE
        """
        return REPEAT.NONE

    def set_repeat(self, mode):
        """
        Sets the repeat mode of the player

        Arguments:
        - `mode`: REPEAT.NONE, REPEAT.TRACK, or REPEAT.ALL
        """
        raise NotImplementedError()

    def get_shuffle(self):
        """
        Gets the shuffle mode of the player

        Returns True if the playlist is shuffle, or False otherwise.

        The default implementation returns False
        """
        return False

    def set_shuffle(self, shuffle):
        """
        Set whether the tracks in track list should be played randomly.

        Arguments:
        - `shuffle`: boolean, True if shuffle
        """
        raise NotImplementedError()

    def play(self):
        """
        Start/continue playing the current track
        """
        raise NotImplementedError()

    def pause(self):
        """
        Pause the current track
        """
        raise NotImplementedError()

    def stop(self):
        """
        Stop playing.
        """
        raise NotImplementedError()

    def prev(self):
        """
        Play the previous track.
        """
        raise NotImplementedError()

    def next(self):
        """
        Play the next track.
        """
        raise NotImplementedError()

    def set_position(self, pos):
        """
        Seek to the given position.

        Arguments:
        - `pos`: Seek time in millisecond
        """
        raise NotImplementedError()

    def get_volume(self):
        """
        Gets the volume of the player.

        Return the volume in the range of [0.0, 1.0]
        """
        raise NotImplementedError()

    def set_volume(self, volume):
        """
        Sets the volume of the player.

        Arguments:
        - `volume`: volume in the range of [0.0, 1.0]
        """
        raise NotImplementedError()

    def _setup_timer_status(self, status):
        status_map = {
            STATUS.PAUSED: 'pause',
            STATUS.PLAYING: 'play',
            STATUS.STOPPED: 'stop',
        }
        if self._timer:
            getattr(self._timer, status_map[status])()

    def _setup_timer(self):
        if self._timer is None:
            self._timer = timer.Timer()
            self._setup_timer_status(self._get_cached_status())

    def _get_cached_position(self):
        """
        Get the current position from cached timer if possible
        """
        if self._timer is None:
            self._setup_timer()
            if self._get_cached_status() != STATUS.STOPPED:
                self._timer.time = self.get_position()
        return self._timer.time

    def _get_cached_loop_status(self):
        if self._loop_status is None:
            self._loop_status = self.get_repeat()
        return self._loop_status

    def _get_cached_status(self):
        if self._status is None:
            self._status = self.get_status()
        return self._status

    def _get_cached_metadata(self):
        if self._metadata is None:
            self._metadata = self._make_metadata(self.get_metadata())
        return self._metadata

    def _make_metadata(self, metadata):
        dct = metadata.to_mpris2()
        dct['mpris:trackid'] = self._get_current_trackid()
        return dct

    def _get_current_trackid(self):
        return '/%s' % self._current_trackid

    def _get_cached_caps(self):
        if self._caps is None:
            self._caps = self.get_caps()
        return self._caps

    def _get_cached_shuffle(self):
        if self._shuffle is None:
            self._shuffle = self.get_shuffle()
        return self._shuffle

    @property
    def connected(self):
        return self._connected

    @property
    def object_path(self):
        return self._object_path

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='',
                         out_signature='')
    def Next(self):
        self.next()

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='',
                         out_signature='')
    def Previous(self):
        self.prev()

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='',
                         out_signature='')
    def Pause(self):
        self.pause()

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='',
                         out_signature='')
    def Stop(self):
        self.stop()

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='',
                         out_signature='')
    def Play(self):
        self.play()

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='x',
                         out_signature='')
    def Seek(self, offset):
        pos = self._get_cached_position()
        pos += offset // 1000
        if pos < 0:
            pos = 0
        self.set_position(pos)

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='ox',
                         out_signature='')
    def SetPosition(self, trackid, position):
        if trackid != self._get_current_trackid():
            return
        self.set_position(position // 1000)

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='',
                         out_signature='')
    def PlayPause(self):
        if hasattr(self, 'play_pause'):
            self.play_pause()
        else:
            status = self._get_cached_status()
            if status == STATUS.PLAYING:
                self.pause()
            else:
                self.play()

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='s',
                         out_signature='')
    def OpenUri(self, uri):
        self.open_uri(uri)

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='s',
                   writeable=False)
    def PlaybackStatus(self):
        status_map = {
            STATUS.PLAYING: 'Playing',
            STATUS.PAUSED: 'Paused',
            STATUS.STOPPED: 'Stopped',
        }
        return status_map[self._get_cached_status()]

    @PlaybackStatus.setter
    def PlaybackStatus(self, status):
        self._status = status

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='s')
    def LoopStatus(self):
        status_map = {
            REPEAT.NONE: 'None',
            REPEAT.ALL: 'Playlist',
            REPEAT.TRACK: 'Track',
        }
        return status_map[self._get_cached_loop_status()]

    @LoopStatus.setter
    def LoopStatus(self, loop_status):
        self._loop_status = loop_status

    @LoopStatus.dbus_setter
    def LoopStatus(self, loop_status):
        status_map = {
            'None': REPEAT.NONE,
            'Playlist': REPEAT.ALL,
            'Track': REPEAT.TRACK,
        }
        if loop_status not in status_map:
            raise ValueError('Unknown loop status ' + loop_status)
        self.set_repeat(status_map[loop_status])

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='d')
    def Rate(self):
        return 1.0

    @Rate.setter
    def Rate(self, rate):
        pass

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='b')
    def Shuffle(self):
        return self._get_cached_shuffle()

    @Shuffle.setter
    def Shuffle(self, shuffle):
        self._shuffle = shuffle

    @Shuffle.dbus_setter
    def Shuffle(self, shuffle):
        self.set_shuffle(shuffle)

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='a{sv}',
                   writeable=False)
    def Metadata(self):
        return self._get_cached_metadata()

    @Metadata.setter
    def Metadata(self, metadata):
        self._metadata = metadata

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='d')
    def Volume(self):
        return self.get_volume()

    @Volume.dbus_setter
    def Volume(self, volume):
        if volume < 0.0:
            volume = 0.0
        if volume > 1.0:
            volume = 1.0
        self.set_volume(volume)

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='x')
    def Position(self):
        return self._get_cached_position()

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='d')
    def MinimumRate(self):
        return 1.0

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='d')
    def MaximumRate(self):
        return 1.0

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='b',
                   writeable=False)
    def CanGoNext(self):
        return CAPS.NEXT in self._get_cached_caps()

    @CanGoNext.setter
    def CanGoNext(self, value):
        pass

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='b',
                   writeable=False)
    def CanGoPrevious(self):
        return CAPS.PREV in self._get_cached_caps()

    @CanGoPrevious.setter
    def CanGoPrevious(self, value):
        pass

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='b',
                   writeable=False)
    def CanPlay(self):
        return CAPS.PLAY in self._get_cached_caps()

    @CanPlay.setter
    def CanPlay(self, value):
        pass

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='b',
                   writeable=False)
    def CanPause(self):
        return CAPS.PAUSE in self._get_cached_caps()

    @CanPause.setter
    def CanPause(self, value):
        pass

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='b',
                   writeable=False)
    def CanSeek(self):
        return CAPS.SEEK in self._get_cached_caps()

    @CanSeek.setter
    def CanSeek(self, value):
        pass

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='b')
    def CanControl(self):
        return True

    @dbus.service.signal(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         signature='x')
    def Seeked(self, position):
        pass

    def track_changed(self, metadata=None):
        self._current_trackid += 1
        if self._timer is not None:
            self._timer.time = self.get_position()
        if metadata is None:
            metadata = self.get_metadata()
        self.Metadata = self._make_metadata(metadata)

    def status_changed(self):
        """
        Notify that the playing status has been changed.
        """
        status = self.get_status()
        self._setup_timer_status(status)
        self.PlaybackStatus = status

    def repeat_changed(self):
        """
        Notify the repeat mode has been changed.
        """
        self.LoopStatus = self.get_repeat()

    def shuffle_changed(self):
        """
        Notify the shuffle mode has been changed.
        """
        self.Shuffle = self.get_shuffle()

    def caps_changed(self):
        """
        Notify the capability of the player has been changed.
        """
        orig_caps = self._caps
        self._caps = self.get_caps()
        if orig_caps is not None:
            caps_map = {
                CAPS.NEXT: 'CanGoNext',
                CAPS.PREV: 'CanGoPrevious',
                CAPS.PLAY: 'CanPlay',
                CAPS.PAUSE: 'CanPause',
                CAPS.SEEK: 'CanSeek',
            }
            for cap, method in caps_map.items():
                if cap in orig_caps != cap in self._caps:
                    setattr(self, method, cap in self._caps)

    def position_changed(self, position):
        """
        Notify that the position has been changed
        """
        if self._timer is not None:
            self._timer.time = position
        self.Seeked(position * 1000)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#
import logging

import dbus

from osdlyrics import PACKAGE_VERSION
from osdlyrics.app import AlreadyRunningException, App
from osdlyrics.consts import (CONFIG_BUS_NAME, DAEMON_BUS_NAME,
                              DAEMON_INTERFACE, DAEMON_MPRIS2_NAME,
                              DAEMON_OBJECT_PATH, MPRIS2_OBJECT_PATH)
from osdlyrics.metadata import Metadata

import lyrics
import lyricsource
import player

logging.basicConfig(level=logging.WARNING)


class InvalidClientNameException(Exception):
    """ The client bus name in Hello is invalid
    """

    def __init__(self, name):
        """

        Arguments:
        - `name`: The invalid client bus name
        """
        super().__init__('Client bus name %s is invalid' % name)


class MainApp(App):
    def __init__(self, ):
        super().__init__('Daemon', False)
        self._player = player.PlayerSupport(self.connection)
        self._lyrics = lyrics.LyricsService(self.connection)
        self._connect_metadata_signal()
        self._activate_config()
        self.request_bus_name(DAEMON_MPRIS2_NAME)
        self._daemon_object = DaemonObject(self)
        self._lyricsource = lyricsource.LyricSource(self.connection)
        self._lyrics.set_current_metadata(Metadata.from_dict(
            self._player.current_player.Metadata))

    def _connect_metadata_signal(self, ):
        self._mpris_proxy = self.connection.get_object(DAEMON_BUS_NAME,
                                                       MPRIS2_OBJECT_PATH)
        self._metadata_signal = self._mpris_proxy.connect_to_signal(
            'PropertiesChanged', self._player_properties_changed)

    def _activate_config(self, ):
        try:
            self.connection.activate_name_owner(CONFIG_BUS_NAME)
        except Exception:
            logging.error("Cannot activate config service")

    def _player_properties_changed(self, iface, changed, invalidated):
        if 'Metadata' in changed:
            self._lyrics.set_current_metadata(Metadata.from_dict(
                changed['Metadata']))


def is_valid_client_bus_name(name):
    """Check if a client bus name is valid.

    A client bus name is valid if it starts with `org.osdlyrics.Client.`

    Arguments:
    - `name`: The bus name of client
    """
    return name.startswith('org.osdlyrics.Client.')


class DaemonObject(dbus.service.Object):
    """ DBus Object implementing org.osdlyrics.Daemon
    """

    def __init__(self, app):
        super().__init__(conn=app.connection,
                         object_path=DAEMON_OBJECT_PATH)
        self._watch_clients = {}
        self._app = app

    @dbus.service.method(dbus_interface=DAEMON_INTERFACE,
                         in_signature='s',
                         out_signature='')
    def Hello(self, client_bus_name):
        logging.info('A new client connected: %s', client_bus_name)
        if is_valid_client_bus_name(client_bus_name):
            if client_bus_name not in self._watch_clients:
                self._watch_clients[client_bus_name] = \
                    self.connection.watch_name_owner(
                        client_bus_name,
                        lambda owner: self._client_owner_changed(
                            client_bus_name, owner))
        else:
            raise InvalidClientNameException(client_bus_name)

    @dbus.service.method(dbus_interface=DAEMON_INTERFACE,
                         in_signature='',
                         out_signature='s')
    def GetVersion(self):
        return PACKAGE_VERSION

    @dbus.service.method(dbus_interface=DAEMON_INTERFACE,
                         in_signature='',
                         out_signature='')
    def Quit(self):
        self._app.quit()

    def _client_owner_changed(self, name, owner):
        if owner == '':
            logging.info('Client %s disconnected', name)
            if name in self._watch_clients:
                self._watch_clients[name].cancel()
                del self._watch_clients[name]
            if not self._watch_clients:
                logging.info('All client disconnected, quit the daemon')
                self._app.quit()


def main():
    try:
        app = MainApp()
        app.run()
    except AlreadyRunningException:
        print('OSD Lyrics is running')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

from error import BadRequestError

__all__ = (
    'baseparam',
    'param_int',
    'param_str',
    'param_enum',
    'param_set',
    'validate_params',
)


class baseparam:
    def __init__(self, optional=False):
        self.optional = optional

    def validate(self, value):
        raise NotImplementedError()


class param_int(baseparam):
    def __init__(self, max=None, min=None, optional=False):
        baseparam.__init__(self, optional)
        self._max = max
        self._min = min

    def validate(self, value):
        try:
            v = int(value)
            if (self._max is None or self._max <= v) and \
                    (self._min is None or self._min >= v):
                return True, v
            else:
                return False, v
        except Exception:
            return False, value


class param_str(baseparam):
    def __init__(self, nonempty=False, optional=False):
        baseparam.__init__(self, optional)
        self._nonempty = nonempty

    def validate(self, value):
        return True if not self._nonempty else isinstance(value, str), value


class param_enum(baseparam):
    def __init__(self, valid_values, optional=False):
        baseparam.__init__(self, optional)
        self._valid_values = valid_values

    def validate(self, value):
        if value not in self._valid_values:
            return False, ''
        try:
            v = self._valid_values[value]
            return True, v
        except Exception:
            return True, value


class param_set(baseparam):
    def __init__(self, valid_values, optional=False):
        baseparam.__init__(self, optional)
        self._valid_values = valid_values

    def validate(self, value):
        ret = set()
        for k in value.split(','):
            k = k.strip()

            if k not in self._valid_values:
                return False, ret
            try:
                v = self._valid_values[k]
                ret.add(v)
            except Exception:
                ret.add(k)
        return True, ret


def validate_params(param_def):
    def dec(func):
        def dec_func(handler, params, *args, **kargs):
            valid_params = {}
            for k, v in params.items():
                if k in param_def:
                    valid, value = param_def[k].validate(v)
                    if not valid:
                        raise BadRequestError('query "%s=%s" is invalid' % (k, v))
                    v = value
                valid_params[k] = v
            for k, v in param_def.items():
                if not v.optional and k not in params:
                    raise BadRequestError('missing "%s" in query' % k)
            return func(handler, valid_params, *args, **kargs)
        return dec_func
    return dec



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

import logging

import dbus

from .consts import CONFIG_BUS_NAME, CONFIG_OBJECT_PATH

CONFIG_INTERFACE = 'org.osdlyrics.Config'


class Config:
    """ Helper class to retrive configs from OSD Lyrics through DBus

    It provides a set of get_<type> functions to get config values with default
    values. If the value with given key exists, return the value. Otherwise the value
    to the key is set to the default value, and returns it. If no default value
    specified and the key does not exist, raise an exception.

    Values can be monitored by connect_change function.
    """

    def __init__(self, conn, follow_name_owner_changes=True):
        """
        Arguments:
        - `conn`: DBus connection
        """
        self._conn = conn
        self._proxy = conn.get_object(CONFIG_BUS_NAME,
                                      CONFIG_OBJECT_PATH,
                                      follow_name_owner_changes=follow_name_owner_changes)
        self._proxy = dbus.Interface(self._proxy,
                                     CONFIG_INTERFACE)
        self._signals = {}
        self._proxy.connect_to_signal('ValueChanged',
                                      self._value_changed_cb)

    def get_bool(self, key, default=None):
        try:
            return self._proxy.GetBool(key)
        except Exception as e:
            if default is not None:
                try:
                    self._proxy.SetBool(key, default)
                except Exception:
                    pass
                return default
            raise e

    def set_bool(self, key, value):
        self._proxy.SetBool(key, value)

    def get_int(self, key, default=None):
        try:
            return self._proxy.GetInt(key)
        except Exception as e:
            if default is not None:
                try:
                    self._proxy.SetInt(key, default)
                except Exception:
                    pass
                return default
            raise e

    def set_int(self, key, value):
        self._proxy.SetInt(key, value)

    def get_double(self, key, default=None):
        try:
            return self._proxy.GetDouble(key)
        except Exception as e:
            if default is not None:
                try:
                    self._proxy.SetBool(key, default)
                except Exception:
                    pass
                return default
            raise e

    def set_double(self, key, value):
        self._proxy.SetDouble(key, value)

    def get_string(self, key, default=None):
        try:
            return self._proxy.GetString(key)
        except Exception as e:
            if default is not None:
                try:
                    self._proxy.SetString(key, default)
                except Exception:
                    pass
                return default
            raise e

    def set_string(self, key, value):
        self._proxy.SetString(key, value)

    def get_string_list(self, key, default=None):
        try:
            return self._proxy.GetStringList(key)
        except Exception as e:
            if default is not None:
                try:
                    self._proxy.SetStringList(key, default)
                except Exception:
                    pass
                return default
            raise e

    def set_string_list(self, key, value):
        self._proxy.SetStringList(key, value)

    def connect_change(self, key, func):
        """
        Calls func once a config value with specified key is changed

        The callback function `func` receives one argument, the key that
        has changed
        """
        if not callable(func):
            return
        self._signals.setdefault(key, []).append(func)

    def disconnect_change(self, key=None, func=None):
        if key is None:
            self._signals = {}
            return
        if key in self._signals:
            if func is None:
                del self._signals[key]
            else:
                self._signals[key].remove(func)

    def _value_changed_cb(self, name_list):
        for name in name_list:
            for handler in self._signals.get(name, []):
                handler(name)


def test():
    def value_changed(name):
        typename = name.split('/')[1]
        logging.debug('%s has been changed to %s', name, getattr(config, 'get_' + typename)(name))

    from gi.repository import GLib
    from dbus.mainloop.glib import DBusGMainLoop
    loop = GLib.MainLoop()
    dbus_mainloop = DBusGMainLoop()
    conn = dbus.SessionBus(mainloop=dbus_mainloop)
    config = Config(conn)
    testcase = {'bool': False,
                'int': 123,
                'double': 123.54,
                'string': 'Foobar',
                'string_list': ['Foo', 'bar'],
                }
    for k in testcase:
        config.connect_change('test/' + k, value_changed)
    for k, v in testcase.items():
        getattr(config, 'set_' + k)('test/' + k, v)
    loop.run()


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

import dbus.exceptions


class AccessDeniedError(dbus.exceptions.DBusException):
    def __init__(self, *args):
        dbus.exceptions.DBusException.__init__(self, dbus_error_name='org.osdlyrics.Error.AccessDenied', *args)


class Property:
    """ DBus property class
    """

    def __init__(self, dbus_interface, type_signature,
                 emit_change=True, readable=True, writeable=True,
                 name=None, fget=None, fset=None, dbus_set=None):
        """

        Arguments:
        - `type_signature`: (string) Type signature of this property. This parameter
          is used for introspection only
        - `dbus_interface`: (string) The DBus interface of this property
        - `emit_change`: (boolean or string) Whether to emit change with
                        `PropertiesChanged` D-Bus signal when the property is set.
                        Possible values are boolean value True or False, or a string
                        'invalidates'.
        - `readable`: Whether the property is able to visit with `Get` D-Bus method.
        - `writeable`: Whether the property is able to write with `Set` D-Bus method.
                       A property is writeable only when `writeable` is set to True
                       and a setter function is set.
        """
        self._type_signature = type_signature
        # CAVEAT: python-dbus uses "_dbus_interface" internally (in service.py)
        self._interface = dbus_interface
        self._fset = fset
        self._fget = fget
        self.__name__ = name
        self._dbusset = dbus_set
        if emit_change not in [True, False, 'invalidates']:
            raise ValueError('Value of emit_change must be one of True, False, or \'invalidates\'')
        self._emit_change = emit_change
        self._readable = readable
        self._writeable = writeable

    @property
    def interface(self):
        """ Return the dbus interface of this property
        """
        return self._interface

    @property
    def readable(self):
        return self._readable

    @property
    def writeable(self):
        return self._writeable

    @property
    def emit_change(self):
        return str(self._emit_change).lower()

    @property
    def type_signature(self):
        return self._type_signature

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        if self._fget is None:
            raise AttributeError("unreadable attribute")
        return wrap_dbus_type(self._type_signature, self._fget(obj))

    def __set__(self, obj, value):
        if self._fset is None:
            raise AttributeError("can't set attribute")
        self._set_value(obj, value, self._fset)

    def dbus_set(self, obj, value):
        """
        Set a property through D-Bus.

        This is intended to called by dbusext.service.Object to implement D-Bus
        property features. Don't call it directly.

        If dbus setter is set by `dbus_setter` decorator, the dbus setter will be
        used the handle this request. Otherwise, if `writeable` is set to `True`,
        the general setter set by `setter` decorator will be used.
        """
        if callable(self._dbusset):
            self._set_value(obj, value, self._dbusset)
        elif self.writeable and callable(self._fset):
            self._set_value(obj, value, self._fset)
        else:
            raise AccessDeniedError('Property %s is not writeable' % self.__name__)

    def _set_value(self, obj, value, setter):
        changed = setter(obj, value)
        if not self._emit_change:
            return
        if changed is None or changed:
            changed = True
        else:
            changed = False
        if changed and getattr(self, '__name__', None) and getattr(obj, '_property_set', None):
            obj._property_set(self.__name__, self._emit_change is True)

    def setter(self, fset):
        """
        Sets the setter function of the property. Return the property object itself.

        The setter function should return a boolean value to tell if the value
        is changed. Returning None is considered as True.

        When ``emit_change`` is True in constructor, the property will emit changes
        when value is set. If the owner object has a function named
        ``_property_set``, and the ``__name__`` attribute of the property object is
        set by the class, the function will be invoked to notify the property has
        been set when the setter is called, with the name of property as the first
        argument, and an boolean as the second argument to tell whether the value
        is changed.

        This is usually used as an decorator::

            class A(osdlyrics.dbus.Object):

              @osdlyrics.dbus.property(type='s', dbus_interface='example.property')
              def x(self):
                  return self._x

              @x.setter
              def x(self, value):
                  if self._x != value:
                      self._x = value
                      return True
                  else:
                      return False

              @x.dbus_setter
              def x(self, value):
                  if self._do_something(value):
                      self._x = value
                      return True
                  return False

        By default, the setter is used as both python property setter and dbus
        property setter. If you want a different set-handler to handle dbus property
        set request, use `dbus_setter`.
        """
        self._fset = fset
        return self

    def dbus_setter(self, fset):
        """
        A decorator to create a D-Bus property set handler.

        The new value set through D-Bus interface is passed to this setter. If dbus
        setter not set, the general setter set by `Property.setter` decorator is
        used as D-Bus property set handler.

        Arguments:
        - `fset`: The setter handler. Takes a parameter as the new value, and return
                  a boolean value to indicate whether the property is changed. If
                  None is returned, it is treated as True. See `setter` for example.
        """
        self._dbusset = fset
        return self


DBUS_TYPE_MAP = {
    'y': dbus.Byte,
    'b': dbus.Boolean,
    'n': dbus.Int16,
    'q': dbus.UInt16,
    'i': dbus.Int32,
    'u': dbus.UInt32,
    'x': dbus.Int64,
    't': dbus.UInt64,
    'd': dbus.Double,
    's': dbus.String,
    'o': dbus.ObjectPath,
    'g': dbus.Signature,
}


def wrap_dbus_type(signature, value):
    if signature in DBUS_TYPE_MAP:
        dbustype = DBUS_TYPE_MAP[signature]
        if isinstance(value, dbustype):
            return value
        else:
            return dbustype(value)
    elif signature.startswith('a{'):
        if isinstance(value, dbus.Dictionary):
            return value
        else:
            return dbus.Dictionary(value, signature=signature)
    elif signature.startswith('a'):
        if isinstance(value, dbus.Array):
            return value
        else:
            return dbus.Array(value, signature=signature)
    elif signature.startswith('('):
        if isinstance(value, dbus.Struct):
            return value
        else:
            return dbus.Struct(value, signature=signature)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 Tiger Soldier <tigersoldi@gmail.com>
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

import os
import os.path
import re
import string

ROOTMAKEFILEAM = r"""SUBDIRS = src
"""

MAKEFILEAM = r"""${name}_PYTHON = ${name}.py

${name}dir = $$(pkglibdir)/lyricsources/${name}


servicedir = $$(datadir)/dbus-1/services
service_in_files = org.osdlyrics.LyricSourcePlugin.${name}.service.in
service_DATA = $$(service_in_files:.service.in=.service)

EXTRA_DIST = \
	$$(service_in_files) \
	$$(NULL)

$$(service_DATA): $$(service_in_files)
	@sed -e "s|\@pkglibdir\@|$$(pkglibdir)|" -e "s|\@PYTHON\@|$$(PYTHON)|" $$< > $$@

CLEANFILES = \
	org.osdlyrics.LyricSourcePlugin.${name}.service \
	$$(NULL)
"""  # noqa: W191

SERVICE = r"""[D-BUS Service]
Name=org.osdlyrics.LyricSourcePlugin.${name}
Exec=@PYTHON@ @pkglibdir@/lyricsources/${name}/${name}.py
"""

PYTHON = r"""# -*- coding: utf-8 -*-

import http.client


class ${capsname}Source(BaseLyricSourcePlugin):
    def __init__(self):
        super().__init__(id='${name}', name='${name}')

    def do_search(self, metadata):
        # type: (osdlyrics.metadata.Metadata) -> List[SearchResult]
        # you can make use of utils.http_download
        #
        # example:
        status, content = http_download(url='http://foo.bar/foobar',
                                        params={'param1': 'foo', 'param2': 'bar'},
                                        proxy=get_proxy_settings(config=self.config_proxy))
        if status < 200 or status >= 400:
            raise http.client.HTTPException(status, '')

        # now do something with content
        return [SearchResult(title='title',
                             artist='artist',
                             album='album',
                             sourceid=self.id,
                             downloadinfo='http://foo.bar/download?id=1')]

    def do_download(self, downloadinfo):
        # type: (Any) -> bytes
        # `downloadinfo` is what you set in SearchResult
        status, content = http_download(url=downloadinfo,
                                        proxy=get_proxy_settings(self.config_proxy))
        if status < 200 or status >= 400:
            raise http.client.HTTPException(status, '')
        return content


if __name__ == '__main__':
    ${name} = ${capsname}Source()
    ${name}._app.run()
"""  # noqa: E101


def input_name():
    prompt = 'Input the lyric source name with only lower-case alphabets and numbers:\n'
    while True:
        name = input(prompt).strip().lower()
        if not re.match(r'[a-z][a-z0-9]*$', name):
            prompt = 'Invalid name. Name must contain only lower-case alphabets and numbers.\nName:'
        else:
            break
    return name


def input_boolean(prompt, default_value):
    prompt += ' [Y/n]?' if default_value is True else ' [y/N]?'
    value = input(prompt)
    if value.lower() == 'y':
        return True
    elif value.lower() == 'n':
        return False
    else:
        return default_value is True


def create_file(template, path, name, params):
    content = string.Template(template).substitute(params)
    f = open(os.path.join(path, name), 'w')
    f.write(content)
    f.close()


def main():
    name = input_name()
    have_am = input_boolean('Generate Makefile.am', True)
    have_subdir = input_boolean('Create source files in src subdirectory', False)
    rootpath = name
    srcpath = name if not have_subdir else name + '/src'
    if not os.path.isdir(srcpath):
        os.makedirs(srcpath)
    params = {
        'name': name,
        'capsname': name.capitalize()
    }
    create_file(PYTHON, srcpath, name + '.py', params)
    create_file(SERVICE, srcpath, 'org.osdlyrics.LyricSourcePlugin.' + name + '.service.in', params)
    if have_am:
        create_file(MAKEFILEAM, srcpath, 'Makefile.am', params)
        if have_subdir:
            create_file(ROOTMAKEFILEAM, rootpath, 'Makefile.am', params)
    print('Done')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#
import logging

import dbus
import dbus.service
import dbus.types

from osdlyrics.consts import (DAEMON_MPRIS2_NAME, MPRIS2_OBJECT_PATH,
                              MPRIS2_PLAYER_INTERFACE, MPRIS2_PREFIX)
from osdlyrics.metadata import Metadata
from osdlyrics.player_proxy import (CAPS, REPEAT, STATUS, BasePlayer,
                                    BasePlayerProxy, PlayerInfo)

# These constants map flags/enums from MPRIS2-specific values to OSDLyrics values.
CAPS_MAP = {
    'CanGoNext': CAPS.NEXT,
    'CanGoPrevious': CAPS.PREV,
    'CanPlay': CAPS.PLAY,
    'CanPause': CAPS.PAUSE,
    'CanSeek': CAPS.SEEK,
}


class ProxyObject(BasePlayerProxy):
    """ The DBus object for MPRIS2 player proxy
    """

    def __init__(self):
        """
        """
        super().__init__('Mpris2')

    def _get_player_from_bus_names(self, names):
        """ Returns list of `PlayerInfo` objects according to names.

        The bus names in names with prefix of MPRIS2_PREFIX will be treated as MPRIS2
        players. The suffix of these names will be treated as player name

        Arguments:
        - `names`: list of bus names
        """
        return [
            PlayerInfo.from_name(name[len(MPRIS2_PREFIX):])
            for name in names
            if name.startswith(MPRIS2_PREFIX) and name != DAEMON_MPRIS2_NAME
        ]

    def do_list_active_players(self):
        return self._get_player_from_bus_names(map(str, self.connection.list_names()))

    def do_list_supported_players(self):
        return self.do_list_activatable_players()

    def do_list_activatable_players(self):
        players = self._get_player_from_bus_names(map(str, self.connection.list_activatable_names()))
        return players

    def do_connect_player(self, player_name):
        player = Mpris2Player(self, player_name)
        return player


class Mpris2Player(BasePlayer):
    def __init__(self, proxy, player_name):
        super().__init__(proxy, player_name)
        self._properties_changed_signal = None
        self._seeked_signal = None
        self._name_watch = None
        try:
            mpris2_object_path = MPRIS2_PREFIX + player_name
            self._player = dbus.Interface(
                self.connection.get_object(mpris2_object_path,
                                           MPRIS2_OBJECT_PATH),
                MPRIS2_PLAYER_INTERFACE)
            self._player_prop = dbus.Interface(
                self.connection.get_object(mpris2_object_path,
                                           MPRIS2_OBJECT_PATH),
                dbus.PROPERTIES_IFACE)
            self._properties_changed_signal = (
                self._player_prop.connect_to_signal(
                    'PropertiesChanged', self._player_properties_changed))
            self._seeked_signal = self._player.connect_to_signal(
                'Seeked', self._player_seeked)
            self._name_watch = self.connection.watch_name_owner(
                mpris2_object_path, self._name_lost)
        except Exception:
            self.disconnect()

    def _name_lost(self, name):
        if name:
            return
        self.disconnect()

    def disconnect(self):
        if self._properties_changed_signal:
            self._properties_changed_signal.remove()
            self._properties_changed_signal = None
        if self._name_watch:
            self._name_watch.cancel()
            self._name_watch = None
        self._player = None
        self._player_prop = None
        BasePlayer.disconnect(self)

    def _player_properties_changed(self, iface, changed, invalidated):
        caps_props = ['CanGoNext', 'CanGoPrevious', 'CanPlay', 'CanPause', 'CanSeek']
        prop_map = {'PlaybackStatus': 'status_changed',
                    'LoopStatus': 'repeat_changed',
                    'Shuffle': 'shuffle_changed',
                    'Metadata': 'track_changed',
                    }
        # status_props = ['PlaybackStatus', 'LoopStatus', 'Shuffle']
        logging.debug('Status changed: %s', changed)
        for caps in caps_props:
            if caps in changed:
                self.caps_changed()
                break
        for prop_name, method in prop_map.items():
            if prop_name in changed:
                getattr(self, method)()

    def _player_seeked(self, position):
        self.position_changed(position // 1000)

    @property
    def object_path(self):
        return self._object_path

    @property
    def connected(self):
        return self._connected

    def next(self):
        self._player.Next()

    def prev(self):
        self._player.Previous()

    def pause(self):
        self._player.Pause()

    def stop(self):
        self._player.Stop()

    def play(self):
        self._player.Play()

    def set_repeat(self, repeat):
        try:
            if repeat == REPEAT.TRACK:
                self._player_prop.Set(MPRIS2_PLAYER_INTERFACE, 'LoopStatus',
                                      'Track')
            elif repeat == REPEAT.ALL:
                self._player_prop.Set(MPRIS2_PLAYER_INTERFACE, 'LoopStatus',
                                      'Playlist')
            else:
                self._player_prop.Set(MPRIS2_PLAYER_INTERFACE, 'LoopStatus',
                                      'None')
        except Exception:
            pass

    def get_status(self):
        playback_dict = {'Playing': STATUS.PLAYING,
                         'Paused': STATUS.PAUSED,
                         'Stopped': STATUS.STOPPED}
        try:
            return playback_dict[self._player_prop.Get(MPRIS2_PLAYER_INTERFACE,
                                                       'PlaybackStatus')]
        except Exception as e:
            logging.error('Failed to get status: %s', e)
            return STATUS.PLAYING

    def get_repeat(self):
        repeat_dict = {'None': REPEAT.NONE,
                       'Track': REPEAT.TRACK,
                       'Playlist': REPEAT.ALL}
        try:
            return repeat_dict[self._player_prop.Get(MPRIS2_PLAYER_INTERFACE,
                                                     'LoopStatus')]
        except Exception:
            return REPEAT.NONE

    def get_shuffle(self):
        try:
            return bool(self._player_prop.Get(MPRIS2_PLAYER_INTERFACE,
                                              'Shuffle'))
        except Exception:
            return False

    def get_metadata(self):
        metadata = self._player_prop.Get(MPRIS2_PLAYER_INTERFACE, 'Metadata')
        return Metadata.from_mpris2(metadata)

    def get_caps(self):
        caps = set()
        for k, cap in CAPS_MAP.items():
            if self._player_prop.Get(MPRIS2_PLAYER_INTERFACE, k):
                caps.add(cap)
        return caps

    def set_volume(self, volume):
        self._player_prop.Set(MPRIS2_PLAYER_INTERFACE, 'Volume', volume)

    def get_volume(self):
        return self._player_prop.Get(MPRIS2_PLAYER_INTERFACE, 'Volume')

    def set_position(self, time_in_mili):
        track_id = self._player_prop.Get(
            MPRIS2_PLAYER_INTERFACE, 'Metadata')['mpris:trackid']
        self._player.SetPosition(track_id, time_in_mili * 1000)

    def get_position(self):
        return self._player_prop.Get(MPRIS2_PLAYER_INTERFACE, 'Position') // 1000


def run():
    mpris2 = ProxyObject()
    mpris2.run()


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012 Tiger Soldier <tigersoldi@gmail.com>
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

from urllib.parse import unquote

# import urlparse
import gettext
import html.parser
import http.client
import re

from osdlyrics.lyricsource import BaseLyricSourcePlugin, SearchResult
from osdlyrics.utils import get_proxy_settings, http_download

_ = gettext.gettext

S4S_HOST = 'https://www.rentanadviser.com'
S4S_SEARCH_PATH = '/en/subtitles/subtitles4songs.aspx'
S4S_SUBTITLE_PATH = '/en/subtitles/getsubtitle.aspx'
# Regexes to filter webpage content (search results page, and lyric page).
S4S_SEARCH_RESULT_PATTERN = re.compile(r'<a [^<]*?href="getsubtitle\.aspx(\?artist=.*?song=.*?)">', re.DOTALL)
S4S_LRC_PATTERN = re.compile(r'<span id="ctl00_ContentPlaceHolder1_lbllyrics"><h3>(.*?)<\/h3>(.*?)<\/span>', re.DOTALL)
TITLE_PATTERN = re.compile(r'&song=(.*$)')
ARTIST_PATTERN = re.compile(r'artist=(.*?)&')
BRAND_PATTERN = r"^.*RentAnAdviser\.com.*$"

gettext.bindtextdomain('osdlyrics')
gettext.textdomain('osdlyrics')


class Subtitles4songsSource(BaseLyricSourcePlugin):
    """ Lyric source from https://www.rentanadviser.com/en/subtitles/subtitles4songs.aspx
    """

    def __init__(self):
        super().__init__(id='subtitles4songs', name=_('subtitles4songs'))

    def do_search(self, metadata):
        # Preparing keywords for search.
        keys = []
        if metadata.artist:
            keys.append(metadata.artist)
        if metadata.title:
            keys.append(metadata.title)
        # Joining search terms.
        urlkey = '+'.join(keys).replace(' ', '+')
        # Building the URL.
        url = S4S_HOST + S4S_SEARCH_PATH
        # Request the HTTP page, storing its status and content.
        status, content = http_download(url=url,
                                        params={'q': urlkey},
                                        proxy=get_proxy_settings(self.config_proxy))
        # Checking against HTTP response codes.
        if status < 200 or status >= 400:
            raise http.client.HTTPException(status, '')
        # Filter all occurences of links belonging to search results.
        matches = S4S_SEARCH_RESULT_PATTERN.findall(content.decode('utf8'))
        # Populate osdlyrics' search results for the user to choose from.
        result = []
        if matches:
            for match in matches:
                # Decompose the HTML matches into title, artist and url.
                title = unquote(TITLE_PATTERN.search(match).group(1))
                artist = unquote(ARTIST_PATTERN.search(match).group(1))
                # Build a download URL for this match.
                url = S4S_HOST + S4S_SUBTITLE_PATH + match + "&type=lrc"
                # Add the match metadata for the window with search results.
                if url is not None:
                    result.append(SearchResult(title=title,
                                               artist=artist,
                                               sourceid=self.id,
                                               downloadinfo=url))
        return result

    def do_download(self, downloadinfo):
        status, content = http_download(downloadinfo,
                                        proxy=get_proxy_settings(self.config_proxy))
        # Checking against HTTP response codes.
        if status < 200 or status >= 400:
            raise http.client.HTTPException(status)
        # Checking the presence of a HTTP payload.
        if content:
            content = S4S_LRC_PATTERN.search(content.decode('utf8')).group(2)
            # Replacing html breaks with new lines.
            content = re.sub('<br />', "\n", content)
            # Debrand the lyrics.
            content = re.sub(BRAND_PATTERN, " ", content, 0, re.MULTILINE)
        return content.encode('utf-8')


if __name__ == '__main__':
    s4s = Subtitles4songsSource()
    s4s._app.run()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

import re

import dbus.types

__all__ = (
    'AttrToken',
    'TimeToken',
    'StringToken',
    'tokenize',
    'parse_lrc',
)


LINE_PATTERN = re.compile(r'(\[[^\[]*?\])')
TIMESTAMP_PATTERN = re.compile(r'^\[(\d+(:\d+){0,2}(\.\d+)?)\]$')
ATTR_PATTERN = re.compile(r'^\[([\w\d]+):(.*)\]$')


class AttrToken:
    """
    Represents tags with the form of ``[key:value]``
    """

    def __init__(self, key, value):
        self.key = key
        self.value = value

    def __repr__(self):
        return '{%s: %s}' % (self.key, self.value)


class StringToken(object):
    """
    Represents a line of lyric text
    """

    def __init__(self, text):
        self.text = text

    def __repr__(self):
        return '"%s"\n' % self.text


class TimeToken(object):
    """
    Represents tags with the form of ``[h:m:s.ms]``

    The time attribute is the timestamp in milliseconds
    """

    def __init__(self, string):
        parts = string.split(':')
        parts.reverse()
        factor = 1000
        ms = int(float(parts[0]) * factor)
        for s in parts[1:]:
            factor = factor * 60
            ms = ms + factor * int(s)
        self.time = ms

    def __repr__(self):
        return '[%s]' % self.time


def tokenize(content):
    """ Split the content of LRC file into tokens

    Returns a list of tokens

    Arguments:
    - `content`: UTF8 string, the content to be tokenized
    """
    def parse_tag(tag):
        m = TIMESTAMP_PATTERN.match(tag)
        if m:
            return TimeToken(m.group(1))
        m = ATTR_PATTERN.match(tag)
        if m:
            return AttrToken(m.group(1), m.group(2))
        return None

    def tokenize_line(line):
        pos = 0
        tokens = []
        while pos < len(line) and line[pos] == '[':
            has_tag = False
            m = LINE_PATTERN.search(line, pos)
            if m and m.start() == pos:
                tag = m.group()
                token = parse_tag(tag)
                if token:
                    tokens.append(token)
                    has_tag = True
                    pos = m.end()
            if not has_tag:
                break
        tokens.append(StringToken(line[pos:]))
        return tokens

    lines = content.splitlines()
    tokens = []
    for line in lines:
        tokens.extend(tokenize_line(line))
    return tokens


def parse_lrc(content):
    """
    Parse an lrc file

    Arguments:
    - `content`: LRC file content encoded in UTF8

    Return values: attr, lyrics
    - `attr`: A dict represents attributes in LRC file
    - `lyrics`: A list of dict with 3 keys: id, timestamp and text.
      The list is sorted in ascending order by timestamp. Id increases from 0.
    """
    tokens = tokenize(content)
    attrs = {}
    lyrics = []
    timetags = []
    for token in tokens:
        if isinstance(token, AttrToken):
            attrs[token.key] = token.value
        elif isinstance(token, TimeToken):
            timetags.append(token.time)
        else:
            for timestamp in timetags:
                lyrics.append({'timestamp': dbus.types.Int64(timestamp),
                               'text': token.text})
            timetags = []
    lyrics.sort(key=lambda a: a['timestamp'])
    i = 0
    for lyric in lyrics:
        lyric['id'] = dbus.types.UInt32(i)
        i = i + 1
    return attrs, lyrics


def test():
    TEST_CASE1 = \
        """[ti:焔の扉~hearty edition][ar:FictionJunction YUUKA]
[al:焔の扉]
[02:45.59]その日まで
[52.78]
[03:48][35]焔の扉へ
[1:03:56.66][03:14.77]
おわり
"""

    def test_tokenizer():
        tokens = tokenize(TEST_CASE1)
        print(tokens)

    def test_parser():
        attr, lyrics = parse_lrc(TEST_CASE1)
        print(attr)
        for line in lyrics:
            print('%s: %s -> %s' % (line['id'], line['timestamp'], line['text']))

    test_tokenizer()
    test_parser()


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#
import os.path
import urllib.parse
import urllib.request

from .errors import PatternException


def expand_file(pattern, metadata):
    """
    Expands the pattern to a file name according to the infomation of a music

    The following are supported place holder in the pattern:
    - %t: Title of the track. 'title' in metadata
    - %p: Performer (artist) of the music. 'artist' in metadata
    - %a: Album of the music. 'album' in metadata
    - %n: Track number of the music. 'tracknumber' in metadata
    - %f: Filename without extension of the music. 'location' in metadata.
    - %%: The `%' punctuation

    Arguments:
    - `pattern`: The pattern to expand.
    - `metadata`: A dict representing metadata. Useful keys are listed above.

    If the pattern cannot be expand, raise an PatternException. Otherwise
    return the expended pattern.

    >>> from osdlyrics.metadata import Metadata
    >>> from_dict = Metadata.from_dict
    >>> metadata = from_dict({'artist': 'Foo',
    ... 'title': 'Bar',
    ... 'tracknumber': '1',
    ... 'album': 'Album',
    ... 'location': 'file:///%E6%AD%8C%E6%9B%B2/%E7%9A%84/%E5%9C%B0%E5%9D%80.mp3'})
    >>> expand_file('%p - %t', metadata)
    'Foo - Bar'
    >>> expand_file('foobar', metadata)
    'foobar'
    >>> print(expand_file('name is %f :)', metadata))
    name is 地址 :)
    >>> expand_file('%something else', metadata)
    '%something else'
    >>> expand_file('%%a - %%t', metadata)
    '%a - %t'
    >>> expand_file('%%%', metadata)
    '%%'
    >>> expand_file('%n - %a:%p,%t', metadata)
    '1 - Album:Foo,Bar'
    >>> expand_file('%t', from_dict({}))
    Traceback (most recent call last):
        ...
    osdlyrics.errors.PatternException: title not in metadata
    """
    keys = {'t': 'title',
            'p': 'artist',
            'a': 'album',
            'n': 'tracknum',
            }
    start = 0
    parts = []
    while start < len(pattern):
        end = pattern.find('%', start)
        if end > -1:
            parts.append(pattern[start:end])
            has_tag = False
            if end + 1 < len(pattern):
                tag = pattern[end + 1]
                if tag == '%':
                    has_tag = True
                    parts.append('%')
                elif tag == 'f':
                    location = metadata.location
                    if not location:
                        raise PatternException('Location not found in metadata')
                    uri = urllib.parse.urlparse(location)
                    if uri.scheme == '':
                        path = uri.path
                    elif uri.scheme == 'file':
                        path = urllib.request.url2pathname(uri.path)
                    else:
                        raise PatternException('Unsupported file scheme %s' % uri.scheme)
                    basename = os.path.basename(path)
                    root, ext = os.path.splitext(basename)
                    has_tag = True
                    parts.append(root)
                elif tag in keys:
                    value = getattr(metadata, keys[tag])
                    if not value:
                        raise PatternException('%s not in metadata' % keys[tag])
                    if not isinstance(value, str):
                        value = str(value)
                    has_tag = True
                    parts.append(value.replace('/', '_'))
            if has_tag:
                start = end + 2
            else:
                start = end + 1
                parts.append('%')
        else:
            parts.append(pattern[start:])
            break
    return ''.join(parts)


def expand_path(pattern, metadata):
    """
    Expands the pattern to a directory path according to the infomation of a music

    The pattern can be one of the three forms:
    - begin with `/': the path is an absolute path and will not be expanded
    - begin with `~/': the path is an relative path and the `~' wiil be expanded to
      the absolute path of the user's home directory
    - `%': the path will be expanded to the directory of the music file according to
      its URI. ``location`` attribute is used in metadata

    Arguments:
    - `pattern`: The pattern to expand.
    - `metadata`: A dict representing metadata. Useful keys are listed above.

    If the pattern cannot be expand, raise an PatternException. Otherwise
    return the expended pattern.


    >>> from osdlyrics.metadata import Metadata
    >>> from_dict = Metadata.from_dict
    >>> expand_path('%', from_dict({'location': 'file:///tmp/a.lrc'}))
    '/tmp'
    >>> expand_path('/bar', from_dict({}))
    '/bar'
    >>> expand_path('%', from_dict({'Title': 'hello'}))
    Traceback (most recent call last):
        ...
    osdlyrics.errors.PatternException: Location not found in metadata
    """
    if pattern == '%':
        location = metadata.location
        if not location:
            raise PatternException('Location not found in metadata')
        uri = urllib.parse.urlparse(location)
        if uri.scheme != 'file':
            raise PatternException('Unsupported scheme: %s' % uri.scheme)
        path = urllib.request.url2pathname(uri.path)
        return os.path.dirname(path)
    path = os.path.expanduser(pattern)
    if not os.path.isabs(path):
        raise PatternException('Path is not absolute: %s' % path)
    return path


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

"""MPD support for OSD Lyrics. Requires MPD >= 0.16 and mpd-python >= 0.3
"""
import logging
import os
import select
import sys

import dbus
import dbus.service
from gi.repository import GLib
try:
    import mpd
    assert hasattr(mpd.MPDClient(), 'send_idle')
except (ImportError, AssertionError):
    logging.error('python-mpd >= 0.3 is required.')
    sys.exit(1)

from osdlyrics.consts import PLAYER_PROXY_INTERFACE
from osdlyrics.metadata import Metadata
from osdlyrics.player_proxy import (CAPS, REPEAT, STATUS, BasePlayer,
                                    BasePlayerProxy, PlayerInfo)
from osdlyrics.timer import Timer
from osdlyrics.utils import cmd_exists

PLAYER_NAME = 'Mpd'
DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 6600


class NoConnectionError(Exception):
    pass


class CommandCallback:
    def __init__(self, command, callback):
        self.command = command
        self.callback = callback

    def call(self, *args):
        if callable(self.callback):
            self.callback(*args)


class Cmds:
    CONFIG = 'config'
    CURRENTSONG = 'currentsong'
    IDLE = 'idle'
    NEXT = 'next'
    NOIDLE = 'noidle'
    PAUSE = 'pause'
    PLAY = 'play'
    PREVIOUS = 'previous'
    RANDOM = 'random'
    REPEAT = 'repeat'
    REPLAY_GAIN_MODE = 'replay_gain_mode'
    REPLAY_GAIN_STATUS = 'replay_gain_status'
    SEEK = 'seek'
    SEEKCUR = 'seekcur'
    SEEKID = 'seekid'
    SETVOL = 'setvol'
    SINGLE = 'single'
    STATUS = 'status'
    STOP = 'stop'


class MpdProxy(BasePlayerProxy):
    def __init__(self):
        super().__init__('Mpd')
        self._player = None
        self._init_address()
        self._client = None
        self._player_info = PlayerInfo(name=PLAYER_NAME,
                                       appname='mpd',
                                       binname='mpd',
                                       cmd='mpd')
        self._player = None
        self._io_watch = None
        self._fetch_queue = []
        self._on_idle = False

    def _init_address(self):
        """
        Initialize the host and port of MPD daemon
        """
        if 'MPD_HOST' in os.environ:
            self._host = os.environ['MPD_HOST']
        else:
            self._host = DEFAULT_HOST
        if 'MPD_PORT' in os.environ and os.environ['MPD_PORT'].isdigit():
            self._port = os.environ['MPD_PORT']
        else:
            self._port = DEFAULT_PORT

    def _connect_mpd(self):
        if self._is_connected():
            return True
        if not self._client:
            self._client = mpd.MPDClient()
        try:
            self._client.connect(self._host, self._port)
        except IOError as e:
            logging.info("Could not connect to '%s': %s", self._host,
                         e.strerror)
            return False
        except mpd.MPDError as e:
            logging.info("Could not connect to '%s': %s", self._host, e)
            return False
        self._io_watch = GLib.io_add_watch(self._client,
                                           GLib.PRIORITY_DEFAULT,
                                           GLib.IOCondition.IN,
                                           self._on_data)
        return True

    def do_list_active_players(self):
        if self._connect_mpd():
            return [self._player_info]
        else:
            return []

    def do_list_supported_players(self):
        return [self._player_info]

    def do_list_activatable_players(self):
        if cmd_exists('mpd'):
            return [self._player_info]
        else:
            return []

    def do_connect_player(self, playername):
        if playername != PLAYER_NAME:
            return None
        if self._player:
            return self._player
        if not self._connect_mpd():
            return None
        self._player = MpdPlayer(self, playername)
        self._start_idle()
        return self._player

    def _on_data(self, client, condition):
        while self._fetch_queue:
            try:
                cmd_item = self._fetch_queue.pop(0)
                logging.debug('fetch cmd: %s', cmd_item.command)
                if not callable(cmd_item.callback):
                    continue
                logging.debug('client pending: %s', self._client._pending)
                retval = getattr(self._client, 'fetch_' + cmd_item.command)()
                cmd_item.call(retval)
            except Exception as e:
                logging.exception(e)
                self._on_disconnect()
            return True

        # no pending data, socket might be closed
        data = os.read(self._client.fileno(), 1024)
        if not data:              # connection closed
            logging.info('connection closed')
            self._on_disconnect()
            return
        else:
            raise RuntimeError('Unexpected data: %s', data)

    def _on_disconnect(self):
        if self._io_watch:
            GLib.remove_source(self._io_watch)
            self._io_watch = None
            self._client.disconnect()
            self._player.disconnect()
            self._player = None
        self._fetch_queue = []
        self._on_idle = False

    def _is_connected(self):
        return True if self._io_watch else False

    def send_command(self, command, callback, *args):
        if not self._is_connected:
            raise NoConnectionError()
        on_idle = self._is_on_idle()
        logging.debug('send %s %s', command, args)
        logging.debug('on idle: %s', on_idle)
        if on_idle:
            self._stop_idle()
        getattr(self._client, 'send_' + command)(*args)
        if callable(callback):
            self._enqueue_callback(command, callback)
        if on_idle:
            self._start_idle()
        logging.debug('cmd queue: %s',
                      [item.command for item in self._fetch_queue])

    def send_command_sync(self, command, callback, *args):
        self.send_command(command, callback, *args)
        logging.debug('send sync to fetch: %s', self._fetch_queue)
        while len(self._fetch_queue) > 1 \
                or (len(self._fetch_queue) == 1
                    and self._fetch_queue[0].command != Cmds.IDLE):
            select.select((self._client,), (), ())
            self._on_data(self._client, GLib.IOCondition.IN)

    def _enqueue_callback(self, command, callback):
        self._fetch_queue.append(CommandCallback(command, callback))

    def _is_on_idle(self):
        return self._fetch_queue and self._fetch_queue[-1].command == Cmds.IDLE

    def _start_idle(self):
        if not self._is_connected() or self._is_on_idle():
            return
        logging.debug('start idle')
        self.send_command(Cmds.IDLE, self._fetch_idle)

    def _stop_idle(self):
        if not self._is_connected() or not self._is_on_idle():
            return
        logging.debug('stop idle')
        self._client.send_noidle()
        self._enqueue_callback(Cmds.NOIDLE, None)

    def _fetch_idle(self, changes):
        if self._player:
            self._player.handle_changes(changes)
        self._start_idle()

    @dbus.service.method(in_signature='',
                         out_signature='a{sv}',
                         dbus_interface=PLAYER_PROXY_INTERFACE)
    def DebugInfo(self):
        ret = {}
        ret['host'] = self._host
        ret['port'] = dbus.UInt32(self._port)
        ret['connected'] = dbus.Boolean(self._is_connected())
        ret['on_idle'] = dbus.Boolean(self._is_on_idle())
        ret['fetch_queue'] = \
            '[' + ','.join([item.command for item in self._fetch_queue]) + ']'
        if self._player:
            ret['player'] = self._player.debug_info()
        return ret


class MpdPlayer(BasePlayer):

    CMD_HANDLERS = {
        Cmds.CURRENTSONG: '_handle_currentsong',
        Cmds.NEXT: None,
        Cmds.PAUSE: None,
        Cmds.PLAY: None,
        Cmds.PREVIOUS: None,
        Cmds.RANDOM: None,
        Cmds.REPEAT: None,
        Cmds.REPLAY_GAIN_MODE: None,
        Cmds.REPLAY_GAIN_STATUS: '_handle_replay_gain_status',
        Cmds.SEEK: None,
        Cmds.SEEKCUR: None,
        Cmds.SEEKID: None,
        Cmds.SETVOL: None,
        Cmds.SINGLE: None,
        Cmds.STATUS: '_handle_status',
        Cmds.STOP: None,
    }

    CHANGE_CMDS = {
        'player': [Cmds.STATUS],
        'options': [Cmds.STATUS],
    }

    STATUS_CHANGE_MAP = {
        'songid': (int, 'track'),
        'playlist': (int, 'track'),
        'repeat': (int, 'repeat'),
        'single': (int, 'repeat'),
        'random': (int, 'shuffle'),
        'state': ('_parse_status', 'status'),
    }

    def __init__(self, proxy, playername):
        super().__init__(proxy, playername)
        self._inited = False
        self._metadata = None
        self._songid = -1
        self._playlist = -1
        self._repeat = None
        self._single = None
        self._random = None
        self._state = None
        self._elapsed = Timer(100)
        self._send_cmd(Cmds.STATUS, sync=True)
        self._inited = True

    def _send_cmd(self, cmd, *args, **kwargs):
        """ Send a cmd. Can use sync=[True|False] to send in a blocking or
        non-blocking way. Default is non-blocking
        """
        sync = False if 'sync' not in kwargs else kwargs['sync']
        if cmd not in self.CMD_HANDLERS:
            raise RuntimeError('Unknown command: %s', cmd)
        handler = self.CMD_HANDLERS[cmd]
        if handler is not None:
            handler = getattr(self, handler)
        else:
            handler = self._handle_nothing
        if sync:
            self.proxy.send_command_sync(cmd, handler, *args)
        else:
            self.proxy.send_command(cmd, handler, *args)

    def _handle_status(self, status):
        logging.debug('status\n%s', status)
        changes = set()
        for prop, handler in self.STATUS_CHANGE_MAP.items():
            if prop not in status:
                value = None
            else:
                func = handler[0]
                if not callable(func):
                    func = getattr(self, func)
                value = func(status[prop])
            if value != getattr(self, '_' + prop):
                logging.debug('prop %s changed to %s', prop, value)
                setattr(self, '_' + prop, value)
                changes.add(handler[1])

        if 'track' in changes:
            if self._songid is None:
                self._metadata = Metadata()
            else:
                self._send_cmd(Cmds.CURRENTSONG, sync=True)

        if 'status' in changes:
            if self._state == STATUS.PAUSED:
                self._elapsed.pause()
            elif self._state == STATUS.PLAYING:
                self._elapsed.play()
            else:
                self._elapsed.stop()
        if self._state == STATUS.STOPPED:
            elapsed = 0
        else:
            elapsed = float(status['elapsed']) * 1000
        if self._elapsed.set_time(elapsed):
            changes.add('position')
        if not self._inited:
            # Initializing, do not emit the change signals
            changes = set()
        for change in changes:
            getattr(self, change + '_changed')()

    def _handle_currentsong(self, metadata):
        logging.debug('currentsong: %s', metadata)
        args = {}
        for key in ('title', 'artist', 'album'):
            if key in metadata:
                args[key] = metadata[key]
        if 'time' in metadata:
            args['length'] = int(metadata['time']) * 1000
        if 'track' in metadata:
            args['tracknum'] = int(metadata['track'].split('/')[0])
        self._metadata = Metadata(**args)

    @staticmethod
    def _parse_status(value):
        status_map = {
            'play': STATUS.PLAYING,
            'pause': STATUS.PAUSED,
            'stop': STATUS.STOPPED,
        }
        if value not in status_map:
            raise RuntimeError('Unknown status ' + value)
        return status_map[value]

    def _handle_replay_gain_status(self, status):
        pass

    def _handle_nothing(self, *args):
        pass

    def handle_changes(self, changes):
        cmds = set()
        for change in changes:
            if change in self.CHANGE_CMDS:
                for cmd in self.CHANGE_CMDS[change]:
                    cmds.add(cmd)
        logging.debug('changes: %s', changes)
        logging.debug('cmds: %s', cmds)
        for cmd in cmds:
            self._send_cmd(cmd)

    def get_status(self):
        return self._state

    def get_metadata(self):
        return self._metadata

    def get_position(self):
        return self._elapsed.time

    def get_caps(self):
        return set([CAPS.PLAY, CAPS.PAUSE, CAPS.NEXT, CAPS.PREV, CAPS.SEEK])

    def get_repeat(self):
        if not self._repeat:
            return REPEAT.NONE
        if not self._single:
            return REPEAT.ALL
        return REPEAT.TRACK

    def set_repeat(self, mode):
        repeat_mode_map = {
            REPEAT.NONE: (0, 0),
            REPEAT.TRACK: (1, 1),
            REPEAT.ALL: (1, 0),
        }
        if mode not in repeat_mode_map:
            raise ValueError('Unknown repeat mode: %s', mode)
        self._repeat = repeat_mode_map[mode][0]
        self._single = repeat_mode_map[mode][1]
        self._send_cmd(Cmds.REPEAT, self._repeat)
        self._send_cmd(Cmds.SINGLE, self._single)

    def get_shuffle(self):
        return bool(self._random)

    def set_shuffle(self, shuffle):
        self._random = 1 if shuffle else 0
        self._send_cmd(Cmds.RANDOM, self._random)

    def play(self):
        if self._state == STATUS.PAUSED:
            self._send_cmd(Cmds.PAUSE, 0)
        elif self._state == STATUS.STOPPED:
            self._send_cmd(Cmds.PLAY)

    def pause(self):
        self._send_cmd(Cmds.PAUSE, 1)

    def stop(self):
        self._send_cmd(Cmds.STOP)

    def prev(self):
        self._send_cmd(Cmds.PREVIOUS)

    def next(self):
        self._send_cmd(Cmds.NEXT)

    def set_position(self, pos):
        self._send_cmd(Cmds.SEEK, self._songid, int(pos // 1000))

    def debug_info(self):
        ret = dbus.Dictionary(signature='sv')
        ret.update({
            'state': self._state,
            'metadata': self._metadata.to_mpris1(),
            'repeat': self._repeat,
            'single': self._single,
            'position': self.get_position()
        })
        return ret


if __name__ == '__main__':
    proxy = MpdProxy()
    proxy.run()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#
import logging

from dbus.exceptions import DBusException
import dbus.service
from gi.repository import GLib

from osdlyrics import PROGRAM_NAME
from osdlyrics.app import App
from osdlyrics.consts import (MPRIS2_OBJECT_PATH, MPRIS2_PLAYER_INTERFACE,
                              PLAYER_PROXY_INTERFACE,
                              PLAYER_PROXY_OBJECT_PATH_PREFIX)
from osdlyrics.dbusext.service import (Object as DBusObject,
                                       property as dbus_property)
import osdlyrics.timer

MPRIS2_ROOT_INTERFACE = 'org.mpris.MediaPlayer2'
PLAYER_INTERFACE = 'org.osdlyrics.Player'
PLAYER_OBJECT_PATH = '/org/osdlyrics/Player'
PLAYER_PROXY_BUS_NAME_PREFIX = 'org.osdlyrics.PlayerProxy.'


class PlayerSupport(dbus.service.Object):
    """ Implement org.osdlyrics.Player Interface
    """

    DETECT_PLAYER_TIMEOUT = 1000

    def __init__(self, conn):
        """
        Arguments:
         - `conn`: DBus connection of the object
        """
        super().__init__(conn=conn, object_path=PLAYER_OBJECT_PATH)
        self._active_player = None
        self._player_proxies = {}
        self._connect_player_proxies()
        self._start_detect_player()
        self._mpris2_player = Mpris2Player(conn)

    def _start_detect_player(self):
        self._detect_timer = GLib.timeout_add(self.DETECT_PLAYER_TIMEOUT,
                                              lambda: not self._detect_player())

    def _detect_player(self):
        """
        Detects active players.

        This is a callback function of a timer. If active player detected, try to
        connect to the player and remove the timer.
        """
        detected = False
        for proxy in self._player_proxies.values():
            try:
                active_players = proxy.ListActivePlayers()
                for player_info in active_players:
                    if self._connect_player(proxy, player_info):
                        detected = True
                        break
                if detected:
                    break
            except Exception:
                pass
        if detected and self._detect_timer:
            GLib.source_remove(self._detect_timer)
            self._detect_timer = None
        return detected

    def _connect_proxy(self, bus_name, activate):
        if not bus_name.startswith(PLAYER_PROXY_BUS_NAME_PREFIX):
            return
        logging.info('Connecting to player proxy %s', bus_name)
        proxy_name = bus_name[len(PLAYER_PROXY_BUS_NAME_PREFIX):]
        if activate:
            try:
                self.connection.activate_name_owner(bus_name)
            except Exception as e:
                logging.warning('Cannot activate proxy %s: %s', bus_name, e)
        self.connection.watch_name_owner(bus_name,
                                         lambda name: self._proxy_name_changed(proxy_name, len(name) == 0))

    def _connect_player_proxies(self):
        """
        Activates all player proxy services
        """
        active_names = map(str, self.connection.list_names())
        for bus_name in active_names:
            self._connect_proxy(bus_name, False)
        activatable_names = map(str, self.connection.list_activatable_names())
        for bus_name in activatable_names:
            self._connect_proxy(bus_name, True)

    def _connect_player(self, proxy, player_info):
        """
        Tries to connect player through proxy

        Return True if connect successful
        """
        try:
            path = proxy.ConnectPlayer(player_info['name'])
        except DBusException as e:
            if e._dbus_error_name != 'org.osdlyrics.Error.ConnectPlayer':
                logging.exception('BugReport')
            return False
        player = self.connection.get_object(proxy.bus_name,
                                            path)
        self._active_player = {'info': player_info,
                               'player': player,
                               'proxy': proxy}
        self._mpris2_player.connect_player(player)
        self.PlayerConnected(player_info)
        return True

    def _player_lost_cb(self, player_name):
        if self._active_player and self._active_player['info']['name'] == player_name:
            logging.info('Player %s lost', player_name)
            self._active_player = None
            self._mpris2_player.disconnect_player()
            self.PlayerLost()
            self._start_detect_player()

    def _proxy_name_changed(self, proxy_name, lost):
        bus_name = PLAYER_PROXY_BUS_NAME_PREFIX + proxy_name
        if not lost:
            logging.info('Get player proxy %s', proxy_name)
            proxy = self.connection.get_object(
                bus_name, PLAYER_PROXY_OBJECT_PATH_PREFIX + proxy_name)
            proxy.connect_to_signal('PlayerLost',
                                    self._player_lost_cb)
            self._player_proxies[proxy_name] = dbus.Interface(
                proxy, PLAYER_PROXY_INTERFACE)
        else:
            if proxy_name not in self._player_proxies:
                return
            logging.info('Player proxy %s lost', proxy_name)
            proxy = self._player_proxies[proxy_name]
            # If current player is provided by the proxy, it is lost.
            if self._active_player and self._active_player['proxy'] == proxy:
                self._player_lost_cb(self._active_player['info']['name'])
                del self._player_proxies[proxy_name]
            # Try to reactivate proxy
            try:
                self.connection.activate_name_owner(bus_name)
            except Exception:
                pass

    @dbus.service.method(dbus_interface=PLAYER_INTERFACE,
                         in_signature='',
                         out_signature='aa{sv}')
    def ListSupportedPlayers(self):
        ret = []
        for proxy in self._player_proxies.values():
            try:
                ret = ret + proxy.ListSupportedPlayers()
            except Exception:
                pass
        return ret

    @dbus.service.method(dbus_interface=PLAYER_INTERFACE,
                         in_signature='',
                         out_signature='aa{sv}')
    def ListActivatablePlayers(self):
        ret = []
        for proxy in self._player_proxies.values():
            try:
                ret = ret + proxy.ListActivatablePlayers()
            except Exception:
                pass
        return ret

    @dbus.service.method(dbus_interface=PLAYER_INTERFACE,
                         in_signature='',
                         out_signature='ba{sv}')
    def GetCurrentPlayer(self):
        if not self._active_player and not self._detect_player():
            return False, {}
        return True, self._active_player['info']

    @dbus.service.signal(dbus_interface=PLAYER_INTERFACE,
                         signature='')
    def PlayerLost(self):
        pass

    @dbus.service.signal(dbus_interface=PLAYER_INTERFACE,
                         signature='a{sv}')
    def PlayerConnected(self, player_info):
        pass

    @property
    def current_player(self):
        return self._mpris2_player


class Mpris2Player(DBusObject):

    def __init__(self, conn):
        super().__init__(conn=conn, object_path=MPRIS2_OBJECT_PATH)
        self._signals = []
        self._player = None
        self._timer = osdlyrics.timer.Timer()
        self._clear_properties()

    def _clear_properties(self):
        self.LoopStatus = 'None'
        self.PlaybackStatus = 'Stopped'
        self.Metadata = dbus.Dictionary(signature='sv')
        self.Shuffle = False
        self._timer.stop()
        self._timer.time = 0

    def connect_player(self, player_proxy):
        if self._player == player_proxy:
            return
        if self._player is not None:
            self.disconnect_player()
        self._player = player_proxy
        self._signals = []
        self._signals.append(self._player.connect_to_signal('Seeked',
                                                            self._seeked_cb))
        self._signals.append(self._player.connect_to_signal('PropertiesChanged',
                                                            self._properties_changed_cb))
        self.PlaybackStatus = self._player.Get(MPRIS2_PLAYER_INTERFACE, 'PlaybackStatus')
        self.LoopStatus = self._player.Get(MPRIS2_PLAYER_INTERFACE, 'LoopStatus')
        self.Shuffle = self._player.Get(MPRIS2_PLAYER_INTERFACE, 'Shuffle')
        self.Metadata = self._player.Get(MPRIS2_PLAYER_INTERFACE, 'Metadata')
        self._setup_timer_status(self._playback_status)
        self._timer.time = self._player.Get(MPRIS2_PLAYER_INTERFACE, 'Position')

    def disconnect_player(self):
        for signal in self._signals:
            signal.remove()
        self._signals = []
        del self._player
        self._player = None
        self._clear_properties()

    def _setup_timer_status(self, status):
        status_map = {
            'Playing': 'play',
            'Paused': 'pause',
            'Stopped': 'stop',
        }
        if status in status_map:
            getattr(self._timer, status_map[status])()

    def _seeked_cb(self, position):
        self._timer.time = position // 1000
        self.Seeked(position)

    def _properties_changed_cb(self, iface, changed, invalidated):
        accepted_properties = set(['PlaybackStatus',
                                   'LoopStatus',
                                   'Rate',
                                   'Shuffle',
                                   'Metadata',
                                   'Volume',
                                   'MinimumRate',
                                   'MaximumRate',
                                   'CanGoNext',
                                   'CanGoPrevious',
                                   'CanPlay',
                                   'CanPause',
                                   'CanSeek',
                                   ])
        for k, v in changed.items():
            if k in accepted_properties:
                setattr(self, k, v)

    ################################################
    # org.mpris.MediaPlayer2.Player interface
    ################################################

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='',
                         out_signature='')
    def Next(self):
        self._player.Next()

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='',
                         out_signature='')
    def Previous(self):
        self._player.Previous()

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='',
                         out_signature='')
    def Pause(self):
        self._player.Pause()

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='',
                         out_signature='')
    def Stop(self):
        self._player.Stop()

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='',
                         out_signature='')
    def Play(self):
        self._player.Play()

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='x',
                         out_signature='')
    def Seek(self, offset):
        self._player.Seek(offset)

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='ox',
                         out_signature='')
    def SetPosition(self, trackid, position):
        self._player.SetPosition(trackid, position)

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='',
                         out_signature='')
    def PlayPause(self):
        self._player.PlayPause()

    @dbus.service.method(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         in_signature='s',
                         out_signature='')
    def OpenUri(self, uri):
        self._player.OpenUri(uri)

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='s',
                   writeable=False)
    def PlaybackStatus(self):
        return self._playback_status

    @PlaybackStatus.setter
    def PlaybackStatus(self, status):
        self._playback_status = status
        self._setup_timer_status(status)

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='s')
    def LoopStatus(self):
        return self._loop_status

    @LoopStatus.setter
    def LoopStatus(self, loop_status):
        self._loop_status = loop_status

    @LoopStatus.dbus_setter
    def LoopStatus(self, loop_status):
        self._player.Set(MPRIS2_PLAYER_INTERFACE, 'LoopStatus', loop_status)

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='d')
    def Rate(self):
        return self._player.Get(MPRIS2_PLAYER_INTERFACE, 'Rate')

    @Rate.setter
    def Rate(self, rate):
        pass

    @Rate.dbus_setter
    def Rate(self, rate):
        self._player.Set(MPRIS2_PLAYER_INTERFACE, 'Rate', rate)

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='b')
    def Shuffle(self):
        return self._shuffle

    @Shuffle.setter
    def Shuffle(self, shuffle):
        self._shuffle = shuffle

    @Shuffle.dbus_setter
    def Shuffle(self, shuffle):
        self._player.Set(MPRIS2_PLAYER_INTERFACE, 'Shuffle', shuffle)

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='a{sv}',
                   writeable=False)
    def Metadata(self):
        return self._metadata

    @Metadata.setter
    def Metadata(self, metadata):
        self._metadata = metadata
        self._timer.time = 0 if not self._player else self._player.Get(
            MPRIS2_PLAYER_INTERFACE, 'Position')

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='d')
    def Volume(self):
        return self._player.Get(MPRIS2_PLAYER_INTERFACE, 'Volume')

    @Volume.setter
    def Volume(self, volume):
        pass

    @Volume.dbus_setter
    def Volume(self, volume):
        return self._player.Set(MPRIS2_PLAYER_INTERFACE, 'Volume', volume)

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='x')
    def Position(self):
        return self._timer.time * 1000

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='d')
    def MinimumRate(self):
        return self._player.Get(MPRIS2_PLAYER_INTERFACE, 'MinimumRate')

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='d')
    def MaximumRate(self):
        return self._player.Get(MPRIS2_PLAYER_INTERFACE, 'MaximumRate')

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='b',
                   writeable=False)
    def CanGoNext(self):
        return self._player.Get(MPRIS2_PLAYER_INTERFACE, 'CanGoNext')

    @CanGoNext.setter
    def CanGoNext(self, value):
        pass

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='b',
                   writeable=False)
    def CanGoPrevious(self):
        return self._player.Get(MPRIS2_PLAYER_INTERFACE, 'CanGoPrevious')

    @CanGoPrevious.setter
    def CanGoPrevious(self, value):
        pass

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='b',
                   writeable=False)
    def CanPlay(self):
        return self._player.Get(MPRIS2_PLAYER_INTERFACE, 'CanPlay')

    @CanPlay.setter
    def CanPlay(self, value):
        pass

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='b',
                   writeable=False)
    def CanPause(self):
        return self._player.Get(MPRIS2_PLAYER_INTERFACE, 'CanPause')

    @CanPause.setter
    def CanPause(self, value):
        pass

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='b',
                   writeable=False)
    def CanSeek(self):
        return self._player.Get(MPRIS2_PLAYER_INTERFACE, 'CanSeek')

    @CanSeek.setter
    def CanSeek(self, value):
        pass

    @dbus_property(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                   type_signature='b')
    def CanControl(self):
        return self._player.Get(MPRIS2_PLAYER_INTERFACE, 'CanControl')

    @dbus.service.signal(dbus_interface=MPRIS2_PLAYER_INTERFACE,
                         signature='x')
    def Seeked(self, position):
        pass

    ################################################
    # org.mpris.MediaPlayer2 interface
    ################################################

    @dbus.service.method(dbus_interface=MPRIS2_ROOT_INTERFACE,
                         in_signature='',
                         out_signature='')
    def Raise(self):
        pass

    @dbus.service.method(dbus_interface=MPRIS2_ROOT_INTERFACE,
                         in_signature='',
                         out_signature='')
    def Quit(self):
        pass

    @dbus_property(dbus_interface=MPRIS2_ROOT_INTERFACE,
                   type_signature='b')
    def CanQuit(self):
        return False

    @dbus_property(dbus_interface=MPRIS2_ROOT_INTERFACE,
                   type_signature='b')
    def Fullscreen(self):
        return False

    @Fullscreen.dbus_setter
    def Fullscreen(self, value):
        pass

    @dbus_property(dbus_interface=MPRIS2_ROOT_INTERFACE,
                   type_signature='b')
    def CanSetFullscreen(self):
        return False

    @dbus_property(dbus_interface=MPRIS2_ROOT_INTERFACE,
                   type_signature='b')
    def CanRaise(self):
        return False

    @dbus_property(dbus_interface=MPRIS2_ROOT_INTERFACE,
                   type_signature='b')
    def HasTrackList(self):
        return False

    @dbus_property(dbus_interface=MPRIS2_ROOT_INTERFACE,
                   type_signature='s')
    def Identity(self):
        return PROGRAM_NAME

    @dbus_property(dbus_interface=MPRIS2_ROOT_INTERFACE,
                   type_signature='s')
    def DesktopEntry(self):
        return 'osdlyrics'

    @dbus_property(dbus_interface=MPRIS2_ROOT_INTERFACE,
                   type_signature='as')
    def SupportedUriSchemes(self):
        return dbus.Array(signature='s')

    @dbus_property(dbus_interface=MPRIS2_ROOT_INTERFACE,
                   type_signature='as')
    def SupportedMimeTypes(self):
        return dbus.Array(signature='s')


def test():
    app = App('osdlyrics')
    mpris2_name = dbus.service.BusName('org.mpris.osdlyrics', app.connection)  # noqa: F841
    player_support = PlayerSupport(app.connection)  # noqa: F841
    app.run()


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#
import io
import os
import os.path
import stat
import sys
import urllib.parse
import urllib.request

import pycurl

__all__ = (
    'cmd_exists',
    'ensure_path',
    'get_config_path',
    'http_download',
    'path2uri',
)

pycurl.global_init(pycurl.GLOBAL_DEFAULT)


class ProxySettings:
    """
    """

    def __init__(self, protocol, host='', port=0, username=None, password=None):
        """
        """
        self.protocol = protocol
        self.host = host
        self.port = port
        self.username = username
        self.password = password


def get_config_path(filename='', expanduser=True):
    """
    Gets the path to save config files

    Arguments:
    - `filename`: (optional string) The filename of config file.
    - `expanduser`: (optional bool) If the leading "~" should be expanded as user's
      home directory

    >>> get_config_path(expanduser=False)
    '~/.config/osdlyrics/'
    >>> get_config_path('osdlyrics.conf', False)
    '~/.config/osdlyrics/osdlyrics.conf'
    """
    path = os.path.join('~/.config/osdlyrics/', filename)
    if expanduser:
        path = os.path.expanduser(path)
    return path


def path2uri(path):
    # type: (Text) -> Text
    r"""
    Converts a path to URI with file sheme.

    If a path does not start with a slash (/), it is considered to be an invalid
    path and returned directly.

    >>> path2uri('/path/to/file')
    'file:///path/to/file'
    >>> path2uri('file:///path/to/file')
    'file:///path/to/file'
    >>> path2uri(u'/path/to/file')
    'file:///path/to/file'
    >>> path2uri('invalid/path')
    'invalid/path'
    >>> path2uri('/\u8def\u5f84/\u6587\u4ef6')
    'file:///%E8%B7%AF%E5%BE%84/%E6%96%87%E4%BB%B6'
    """
    path = os.path.expanduser(path)
    if not os.path.isabs(path):
        return path
    return 'file://' + urllib.request.pathname2url(path)


def get_proxy_settings(config=None, conn=None):
    r"""
    Return proxy settings as a ProxySettings object

    The caller must specify either config or conn.

    Arguments:
     - `config`: A osdlyrics.config.Config object, this object is used to retrive
                 proxy settings. If it is not set, the caller MUST set conn to a
                 valid D-Bus connection to create a Config object
     - `conn`: A D-Bus connection object, this is used when `config` is not
               specified.
    """
    if config is None and conn is None:
        raise ValueError('Either config or conn must be specified')
    if config is None:
        config = config.Config(conn)
    proxy_type = config.get_string('Download/proxy').lower()
    if proxy_type == 'no':
        return ProxySettings(protocol='no')
    if proxy_type == 'manual':
        protocol = config.get_string('Download/proxy-type')
        host = config.get_string('Download/proxy-host')
        port = config.get_int('Download/proxy-port')
        username = config.get_string('Download/proxy-username')
        passwd = config.get_string('Download/proxy-password')
        return ProxySettings(protocol=protocol, host=host, port=port,
                             username=username, password=passwd)
    if proxy_type == 'system':
        return detect_system_proxy()


def detect_system_proxy():
    r"""
    Detects and return system proxy settings.

    Support following proxy settings:
    - Environment variables
    - GNOME 2 (TODO)
    - GNOME 3
    - KDE
    """
    desktop = detect_desktop_shell()
    if desktop == 'gnome' or desktop == 'unity':
        proxy = get_gsettings_proxy()
        if proxy is not None:
            return proxy
    elif desktop == 'kde':
        proxy = get_kde_proxy()
        if proxy is not None:
            return proxy
    return get_envar_proxy()


def get_envar_proxy():
    r"""
    Return proxy settings from environment variable `http_proxy`
    """
    envars = ['http_proxy', 'HTTP_PROXY']
    proxies = [os.environ.get(v) for v in envars]
    for proxy in proxies:
        if proxy:
            if '://' not in proxy:
                proxy = 'http://' + proxy
            parts = urllib.parse.urlparse(proxy)
            if parts.scheme not in ('http', 'socks4', 'socks5', ''):
                continue
            return ProxySettings(protocol=parts.scheme or 'http',
                                 host=parts.hostname,
                                 port=parts.port or 8080,
                                 username=parts.username,
                                 password=parts.password)
    return ProxySettings(protocol='no')


def detect_desktop_shell():
    r"""
    Detect the currently running destop shell.

    Returns: 'gnome', 'unity', 'kde', or 'unknown'
    """
    envar = os.environ.get('DESKTOP_SESSION')
    if envar.startswith('gnome'):
        return 'gnome'
    if envar.startswith('kde'):
        return 'kde'
    if envar.startswith('ubuntu') or envar.startswith('unity'):
        return 'unity'
    return 'unknown'


def get_gsettings_proxy():
    r"""
    Return proxy settings from gsetting, this is used in GNOME 3
    """
    try:
        from gi.repository import Gio
    except ImportError:
        return None
    if not hasattr(Gio, 'Settings'):
        return None
    if 'org.gnome.system.proxy' not in Gio.Settings.list_schemas():
        return None
    settings = Gio.Settings('org.gnome.system.proxy')
    if settings.get_string('mode') != 'manual':
        return ProxySettings(protocol='no')
    protocol_map = {'http': 'http', 'socks5': 'socks'}
    for protocol, key in protocol_map.items():
        settings = Gio.Settings('org.gnome.system.proxy.' + key)
        host = settings.get_string('host').strip()
        port = settings.get_int('port')
        if host == '' or port <= 0:
            continue
        username = ''
        password = ''
        if key == 'http' and settings.get_boolean('use-authentication'):
            username = settings.get_string('authentication-user')
            password = settings.get_string('authentication-password')
        return ProxySettings(protocol=protocol,
                             host=host,
                             port=port,
                             username=username,
                             password=password)
    return ProxySettings(protocol='no')


def get_kde_proxy():
    r"""
    Detect KDE4 proxy settings
    """
    try:
        import PyKDE4.kdecore as kdecore
    except ImportError:
        return None
    config = kdecore.KConfig('kioslaverc', kdecore.KConfig.NoGlobals)
    if not config.hasGroup('Proxy Settings'):
        return None
    group = config.group('Proxy Settings')
    proxy_type, _ = group.readEntry('ProxyType', 0).toInt()
    if proxy_type in [0, 2, 3]:
        # no proxy, PAC or auto detect in KDE settings. We don't support
        # PAC proxy, so treat them as no proxy
        return ProxySettings('no')
    elif proxy_type in [1, 4]:
        for key in ['httpProxy', 'socksProxy']:
            value = str(group.readEntry(key))
            if value.strip():
                # KDE 4.8 uses whitespace to seperate port and hostname
                value = value.replace(' ', ':')
                if '://' not in value:
                    value = 'http://' + value
                parts = urllib.parse.urlparse(value)
                host = parts.hostname
                port = parts.port
                if host is not None and host.strip() != '' and \
                        port is not None and 0 < port < 65536:
                    protocolmap = {'httpProxy': 'http',
                                   'socksProxy': 'socks5'}
                    return ProxySettings(protocolmap[key],
                                         host=host,
                                         port=port)
    return ProxySettings('no')


def http_download(url, port=0, method='GET', params={}, headers={}, timeout=15, proxy=None):
    # type: (Text, int, Text, Union[bytes, Dict[Text, Any]], Dict[Text, Any], int, Any) -> Tuple[int, bytes]
    r"""
    Helper function to download files from website

    This function will apply proxy settings and deal redirections automatically.
    To apply proxy settings, pass an ProxySettings object as the `proxy` parameter.

    If `'User-Agent'` is not set in `headers`, it will be set to `'OSD Lyrics'`.

    Arguments:
     - `url`: The url of the content. Must be a bytes or an ascii-encodable unicode.
     - `port`: (optional) The port.
     - `method`: (optional) The HTTP method to download contents. Available values
                 are `'POST'` or `'GET'`. The default value is `'GET'`.
     - `params`: (optional) The parameters of the request. It is either a dict or bytes.
                 If it is a dict, it will be utf8-urlencoded. If `method` is `'GET'`,
                 `params` will be append to the url as the param part. If `method` is
                 `'POST'`, `params` will be added to request body as post data.
     - `headers`: (optional) A dict of HTTP headers.
     - `proxy`: (optional) A ProxySettings object to sepcify the proxy to use.

    >>> code, content = http_download('http://www.python.org/')
    >>> code
    200
    >>> b'Python' in content
    True
    """
    c = pycurl.Curl()
    buf = io.BytesIO()
    c.setopt(pycurl.NOSIGNAL, 1)
    c.setopt(pycurl.DNS_USE_GLOBAL_CACHE, 0)
    c.setopt(pycurl.FOLLOWLOCATION, 1)
    c.setopt(pycurl.MAXREDIRS, 5)
    c.setopt(pycurl.WRITEFUNCTION, buf.write)
    if isinstance(params, dict):
        params = urllib.parse.urlencode(params)
    if method == 'GET' and params:
        url = url + ('/' if '/' not in url else '') + ('?' if '?' not in url else '&') + params
    elif method == 'POST':
        c.setopt(pycurl.POST, 1)
        if params:
            c.setopt(pycurl.POSTFIELDS, params)
            c.setopt(pycurl.POSTFIELDSIZE, len(params))
    c.setopt(pycurl.URL, url)
    if 0 < port < 65536:
        c.setopt(pycurl.PORT, port)

    real_headers = {'User-Agent': 'OSD Lyrics'}
    real_headers.update(headers)
    curl_headers = ['%s:%s' % (k, v) for k, v in real_headers.items()]
    c.setopt(pycurl.HTTPHEADER, curl_headers)

    if proxy is not None and proxy.protocol != 'no':
        if proxy.username != '' and proxy.username is not None:
            proxyvalue = '%s://%s:%s@%s:%d' % (proxy.protocol,
                                               proxy.username,
                                               proxy.password,
                                               proxy.host,
                                               proxy.port)
        else:
            proxyvalue = '%s://%s:%d' % (proxy.protocol,
                                         proxy.host,
                                         proxy.port)
        c.setopt(pycurl.PROXY, proxyvalue)
    else:
        c.setopt(pycurl.PROXY, '')

    c.perform()
    return c.getinfo(pycurl.HTTP_CODE), buf.getvalue()


def ensure_path(path, ignore_file_name=True):
    """ Create directories if necessary.

    This function tries to create directories for `path`. Unlike `os.makedirs`,
    no error will be raised if the leaf directory exists.

    Arguments:
    - `path`: The path.
    - `ignore_file_name`: (optional) If set to `True`, the path to be create will be
       the directory part of `path` which is get from `os.path.dirname(path)`.
       Otherwise `path` is considered to be path to a directory rather than a file.
       The default value is `True`.
    """
    if ignore_file_name:
        path = os.path.dirname(path)

    if os.path.isdir(path):
        return
    os.makedirs(path)


def find_file_in_dirs(filename, dirs, filter_func=None):
    """
    Find file in specified directories. Return a list of full path of found
    files
    Arguments:
    - `filename`: the filename to find
    - `dirs`: a list of directory paths
    - `filter_func`: a filter function that takes the full path of a file
                     and returns a boolean value. If `False` is returned by the
                     function, the path will not be in the returned list.
                     Default value is None
    """
    ret = []
    for dir in dirs:
        path = os.path.join(dir, filename)
        if os.path.isfile(path) and \
                (not callable(filter_func) or filter_func(path)):
            ret.append(path)
    return ret


def cmd_exists(cmd):
    """
    Check if a command exists.
    >>> cmd_exists('ls')
    True
    >>> cmd_exists('cmd_not_likely_exists')
    False
    """
    cmdfiles = find_file_in_dirs(cmd,
                                 os.environ['PATH'].split(':'),
                                 is_exec_file)
    return len(cmdfiles) > 0


def is_exec_file(filepath):
    """
    Check if a file is executable by current user
    """
    try:
        st = os.stat(filepath)
    except Exception:
        return False
    uid = os.getuid()
    gid = os.getgroups()
    if st.st_uid == uid:
        return (st.st_mode & stat.S_IXUSR) != 0
    if st.st_gid in gid:
        return (st.st_mode & stat.S_IXGRP) != 0
    return (st.st_mode & stat.S_IXOTH) != 0


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#
from datetime import datetime


class Timer:
    """ A timer to account the elapsed playing.
    """

    def __init__(self, accuracy=0):
        """

        Arguments:
        - `accuracy`:The accuracy of the time provided by players.
          If the difference between the time provided by players and the time of
          the timer is greater than accuracy, the time of the timer will be adjust to
          the time of the player.
        """
        self._accuracy = accuracy
        self._started = False
        self._time = 0
        self._begintime = None

    def play(self):
        """
        Starts the timer
        """
        if not self._started:
            self._started = True
            self._begintime = datetime.now()

    def pause(self):
        """
        Stops the timer
        """
        self._time = self.time
        self._begintime = None
        self._started = False

    def stop(self):
        """
        Stops the timer and sets the elapsed time to 0
        """
        self.pause()
        self._time = 0

    @property
    def time(self):
        now = datetime.now()
        if self._started:
            return self._time + int((now - self._begintime).total_seconds() * 1000)
        else:
            return self._time

    @time.setter
    def time(self, value):
        self.set_time(value)

    def set_time(self, value):
        """
        Adjust the time.

        If the time needs to be adjusted, return True.
        """
        time = self.time
        if abs(time - value) > self._accuracy:
            self._time = value
            self._begintime = datetime.now()
            return True
        return False



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
import os
import os.path
import re
import urllib.parse
import urllib.request

import chardet
import dbus
import dbus.service

import osdlyrics
from osdlyrics.app import App
import osdlyrics.config
import osdlyrics.lrc
from osdlyrics.metadata import Metadata
from osdlyrics.pattern import expand_file, expand_path

import lrcdb

LYRICS_INTERFACE = 'org.osdlyrics.Lyrics'
LYRICS_OBJECT_PATH = '/org/osdlyrics/Lyrics'

DEFAULT_FILE_PATTERNS = [
    '%p-%t',
    '%t-%p',
    '%f',
    '%t',
]

DEFAULT_PATH_PATTERNS = [
    '~/.lyrics',
    '%',
]

SUPPORTED_SCHEMES = [
    'file',
    # 'tag',
    'none',
]

DETECT_CHARSET_GUESS_MIN_LEN = 40
DETECT_CHARSET_GUESS_MAX_LEN = 100


class InvalidUriException(Exception):
    """ Exception of invalid uri.
    """

    def __init__(self, uri):
        super().__init__("Invalid URI: %s" % uri)


class CannotLoadLrcException(Exception):
    def __init__(self, uri):
        super().__init__("Cannot load lrc file from %s" % uri)


class CannotSaveLrcException(Exception):
    def __init__(self, uri):
        super().__init__("Cannot save lrc file to %s" % uri)


def metadata_description(metadata):
    if metadata.title is None:
        return '[Unknown]'
    if metadata.artist is None:
        return metadata.title
    return '%s(%s)' % (metadata.title, metadata.artist)


def decode_by_charset(content):
    # type: (bytes) -> Text
    r"""
    Detect the charset encoding of a string and decodes to unicode strings.

    >>> decode_by_charset(u'\u4e2d\u6587'.encode('UTF-8'))
    '\u4e2d\u6587'
    >>> decode_by_charset(u'\u4e2d\u6587'.encode('HZ-GB-2312'))
    '\u4e2d\u6587'
    """
    encoding = chardet.detect(content)['encoding']
    # Sometimes, the content is well encoded but the last few bytes. This is
    # common in the files downloaded by old versions of OSD Lyrics. In this
    # case,chardet may fail to determine what the encoding it is. So we take
    # half of the content of it and try again.
    if not encoding and len(content) > DETECT_CHARSET_GUESS_MIN_LEN:
        logging.warning('Failed to detect encoding, try to decode a part of it')
        content_half = len(content) // 2
        slice_end = min(max(DETECT_CHARSET_GUESS_MIN_LEN, content_half), DETECT_CHARSET_GUESS_MAX_LEN)
        encoding = chardet.detect(content[:slice_end])['encoding']
        logging.warning('guess encoding from part: ' + encoding)
    if not encoding:
        logging.warning('Failed to detect encoding, use utf-8 as fallback')
        encoding = 'utf-8'

    encoding = encoding.lower()
    # When we take half of the content to determine the encoding, chardet may
    # think it be encoded with ascii, however the full content is probably
    # encoded with utf-8. As ascii is an subset of utf-8, decoding an ascii
    # string with utf-8 will always be right.
    if encoding == 'ascii':
        encoding = 'utf-8'
    # Upgrade the Chinese encodings to their extended sets.
    elif encoding in ('gb2312', 'gbk'):
        encoding = 'gb18030'
    elif encoding == 'big5':
        encoding = 'big5hkscs'
    return content.decode(encoding, 'replace')


def is_valid_uri(uri):
    """
    Tell if a URI is valid.

    A valid URI must begin with the schemes defined in SUPPORTED_SCHEMES
    """
    for scheme in SUPPORTED_SCHEMES:
        if uri.startswith(scheme + ':'):
            return True
    return False


def ensure_uri_scheme(uri):
    # type: (Text) -> Text
    """
    Converts a file path to an URI with scheme of "file:", leaving other URI not
    changed

    If the uri doesn't have any scheme, it is considered to be a file path.
    """
    if uri:
        url_parts = urllib.parse.urlparse(uri)
        if not url_parts.scheme:
            uri = osdlyrics.utils.path2uri(uri)
    return uri


def _load_from_file(urlparts):
    """
    Load the content of file from urlparse.ParseResult

    Return the content of the file, or None if error raised.
    """
    path = urllib.request.url2pathname(urlparts.path)
    try:
        with open(path, 'rb') as f:
            return f.read()
    except IOError as e:
        logging.info("Cannot open file %s to read: %s", path, e)
        return None


def load_from_uri(uri):
    # type: (Text) -> Optional[Text]
    """
    Load the content of LRC file from given URI

    If loaded, return the content. If failed, return None.
    """
    URI_LOAD_HANDLERS = {
        'file': _load_from_file,
        'none': lambda uri: b'',
    }

    url_parts = urllib.parse.urlparse(uri)
    content = URI_LOAD_HANDLERS[url_parts.scheme](url_parts)
    if content is None:
        return None
    content = decode_by_charset(content).replace('\0', '')
    return content


def _save_to_file(urlparts, content, create):
    # type: (Any, bytes, bool) -> bool
    """
    Save the content of file to urlparse.ParseResult

    Return True if succeeded
    """
    path = urllib.request.url2pathname(urlparts.path)
    if not create:
        if not os.path.isfile(path):
            logging.warning("Cannot write to file %s: file not exists", path)
            return False
    else:
        dirname = os.path.dirname(path)
        if not os.path.isdir(dirname):
            try:
                os.makedirs(os.path.dirname(path), 0o755)
            except OSError as e:
                logging.warning("Cannot create directories for %s: %s", path, e)
                return False
    try:
        file = open(path, 'wb')
    except IOError as e:
        logging.info("Cannot open file %s to write: %s", path, e)
        return False
    file.write(content)
    return True


def save_to_uri(uri, content, create=True):
    # type: (Text, bytes, bool) -> bool
    """
    Save the content of LRC file to given URI.

    Return True if succeeded, or False if failed.
    """
    URI_SAVE_HANDLERS = {
        'file': _save_to_file,
        'none': lambda urlparts, content, create: True,
    }

    url_parts = urllib.parse.urlparse(uri)
    return URI_SAVE_HANDLERS[url_parts.scheme](url_parts, content, create)


def update_lrc_offset(content, offset):
    r"""
    Replace the offset attributes in the content of LRC file.
    >>> update_lrc_offset('no tag', 100)
    '[offset:100]\nno tag'
    >>> update_lrc_offset('[ti:title]\n[offset:200]\nSome lrc', 100)
    '[ti:title]\n[offset:100]\nSome lrc'
    >>> update_lrc_offset('[ti:title][offset:200]Some lrc\nanother', 100)
    '[ti:title][offset:100]Some lrc\nanother'
    >>> update_lrc_offset('Some [offset:200] lrc', 100)
    '[offset:100]\nSome [offset:200] lrc'
    >>> update_lrc_offset('[[offset:200]] lrc', 100)
    '[offset:100]\n[[offset:200]] lrc'
    """
    search_result = re.search(r'^(\[[^\]]*\])*?\[offset:(.*?)\]',
                              content,
                              re.MULTILINE)
    if search_result is None:
        return '[offset:%s]\n%s' % (offset, content)
    return '%s%s%s' % (content[:search_result.start(2)],
                       offset,
                       content[search_result.end(2):])


class LyricsService(dbus.service.Object):

    def __init__(self, conn):
        super().__init__(conn=conn, object_path=LYRICS_OBJECT_PATH)
        self._db = lrcdb.LrcDb()
        self._config = osdlyrics.config.Config(conn)
        self._metadata = Metadata()

    def find_lrc_from_db(self, metadata):
        uri = self._db.find(metadata)
        if uri == '':
            return 'none:'
        return ensure_uri_scheme(uri)

    def find_lrc_by_pattern(self, metadata):
        return ensure_uri_scheme(self._expand_patterns(metadata))

    def assign_lrc_uri(self, metadata, uri):
        self._db.assign(metadata, uri)
        if metadata == self._metadata:
            self.CurrentLyricsChanged()

    @dbus.service.method(dbus_interface=LYRICS_INTERFACE,
                         in_signature='a{sv}',
                         out_signature='bsa{ss}aa{sv}')
    def GetLyrics(self, metadata):
        ret, uri, content = self.GetRawLyrics(metadata)
        if ret:
            attr, lines = osdlyrics.lrc.parse_lrc(content)
            return ret, uri, attr, lines
        else:
            return ret, uri, {}, []

    @dbus.service.method(dbus_interface=LYRICS_INTERFACE,
                         in_signature='a{sv}',
                         out_signature='bss')
    def GetRawLyrics(self, metadata):
        if isinstance(metadata, dict):
            metadata = Metadata.from_dict(metadata)
        uri = self.find_lrc_from_db(metadata)
        lrc = None
        if uri:
            if uri == 'none:':
                return True, uri, ''
            lrc = load_from_uri(uri)
            if lrc is not None:
                return True, uri, lrc
        uri = self.find_lrc_by_pattern(metadata)
        if uri:
            lrc = load_from_uri(uri)
            if lrc is not None:
                logging.info("LRC for track %s not found in db but found by pattern: %s", metadata_description(metadata), uri)
        if lrc is None:
            logging.info("LRC for track %s not found", metadata_description(metadata))
            return False, '', ''
        else:
            logging.info("LRC for track %s found: %s", metadata_description(metadata), uri)
            return True, uri, lrc

    @dbus.service.method(dbus_interface=LYRICS_INTERFACE,
                         in_signature='',
                         out_signature='bsa{ss}aa{sv}')
    def GetCurrentLyrics(self):
        return self.GetLyrics(self._metadata)

    @dbus.service.method(dbus_interface=LYRICS_INTERFACE,
                         in_signature='',
                         out_signature='bss')
    def GetCurrentRawLyrics(self):
        return self.GetRawLyrics(self._metadata)

    @dbus.service.method(dbus_interface=LYRICS_INTERFACE,
                         in_signature='a{sv}ay',
                         out_signature='s',
                         byte_arrays=True)
    def SetLyricContent(self, metadata, content):
        metadata = Metadata.from_dict(metadata)
        # Remove any existing file association and save the new lyrics content
        # to the configured patterns.
        self._db.delete(metadata)
        uri = self._save_to_patterns(metadata, content.rstrip(b'\0'))
        if uri and metadata == self._metadata:
            self.CurrentLyricsChanged()
        return uri

    @dbus.service.method(dbus_interface=LYRICS_INTERFACE,
                         in_signature='a{sv}s',
                         out_signature='')
    def AssignLyricFile(self, metadata, uri):
        metadata = Metadata.from_dict(metadata)
        self.assign_lrc_uri(metadata, uri)

    @dbus.service.signal(dbus_interface=LYRICS_INTERFACE,
                         signature='')
    def CurrentLyricsChanged(self):
        pass

    @dbus.service.method(dbus_interface=LYRICS_INTERFACE,
                         in_signature='si',
                         out_signature='')
    def SetOffset(self, uri, offset_ms):
        if not is_valid_uri(uri):
            raise InvalidUriException(uri)
        content = load_from_uri(uri)
        if content is None:
            raise CannotLoadLrcException(uri)
        content = update_lrc_offset(content, offset_ms).encode('utf-8')
        if not save_to_uri(uri, content, True):
            raise CannotSaveLrcException(uri)

    def _save_to_patterns(self, metadata, content):
        """ Save content to file expanded from given patterns

        Arguments:
        - `metadata`:
        - `content`:
        """
        file_patterns = self._config.get_string_list('General/lrc-filename',
                                                     DEFAULT_FILE_PATTERNS)
        path_patterns = self._config.get_string_list('General/lrc-path',
                                                     DEFAULT_PATH_PATTERNS)
        for path_pat in path_patterns:
            try:
                path = expand_path(path_pat, metadata)
            except osdlyrics.pattern.PatternException:
                continue
            for file_pat in file_patterns:
                try:
                    filename = expand_file(file_pat, metadata)
                except osdlyrics.pattern.PatternException:
                    continue
                fullpath = os.path.join(path, filename + '.lrc')
                uri = osdlyrics.utils.path2uri(fullpath)
                if save_to_uri(uri, content):
                    return uri
        return ''

    def _expand_patterns(self, metadata):
        file_patterns = self._config.get_string_list('General/lrc-filename',
                                                     DEFAULT_FILE_PATTERNS)
        path_patterns = self._config.get_string_list('General/lrc-path',
                                                     DEFAULT_PATH_PATTERNS)
        for path_pat in path_patterns:
            try:
                path = expand_path(path_pat, metadata)
            except osdlyrics.pattern.PatternException:
                continue
            for file_pat in file_patterns:
                try:
                    filename = expand_file(file_pat, metadata)
                except osdlyrics.pattern.PatternException:
                    continue
                fullpath = os.path.join(path, filename + '.lrc')
                if os.path.isfile(fullpath):
                    return fullpath
        return None

    def set_current_metadata(self, metadata):
        logging.info('Setting current metadata: %s', metadata)
        self._metadata = metadata


def doc_test():
    import doctest
    doctest.testmod()


def test():
    app = App('Lyrics', False)
    lyrics_service = LyricsService(app.connection)  # noqa: F841
    app.run()


if __name__ == '__main__':
    doc_test()
    test()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011  Tiger Soldier
#
# This file is part of OSD Lyrics.
#
# OSD Lyrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# OSD Lyrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OSD Lyrics.  If not, see <https://www.gnu.org/licenses/>.
#

import configparser

import dbus
import dbus.service
from gi.repository import GLib

from osdlyrics.app import App
from osdlyrics.consts import CONFIG_BUS_NAME, CONFIG_OBJECT_PATH
import osdlyrics.errors
import osdlyrics.utils


class MalformedKeyError(osdlyrics.errors.BaseError):
    pass


class ValueNotExistError(osdlyrics.errors.BaseError):
    def __init__(self, key=''):
        super().__init__('Value of key %s does not exist' % key)


class IniConfig(dbus.service.Object):
    """ Implement org.osdlyrics.Config
    """

    def __init__(self,
                 conn,
                 filename=osdlyrics.utils.get_config_path('osdlyrics.conf')):
        super().__init__(conn=conn, object_path=CONFIG_OBJECT_PATH)
        self._conn = conn
        self._confparser = configparser.RawConfigParser()
        osdlyrics.utils.ensure_path(filename)
        self._confparser.read(filename)
        self._filename = filename
        self._save_timer = None
        self._signal_timer = None
        self._changed_signals = {}

    def _split_key(self, key, add_section=True):
        parts = key.split('/')
        if len(parts) != 2:
            raise MalformedKeyError(
                '%s is an invalid key. Keys must be in the form '
                'of "Section/Name"' % key
            )
        if not parts[0] or not parts[1]:
            raise MalformedKeyError(
                'Malformed key "%s". Section or name must not be empty' % key)
        if add_section and not self._confparser.has_section(parts[0]):
            self._confparser.add_section(parts[0])
        return parts[0], parts[1]

    @dbus.service.method(dbus_interface=CONFIG_BUS_NAME,
                         in_signature='s',
                         out_signature='b')
    def GetBool(self, key):
        section, name = self._split_key(key)
        try:
            return self._confparser.getboolean(section, name)
        except (configparser.NoSectionError, configparser.NoOptionError):
            raise ValueNotExistError(key)

    @dbus.service.method(dbus_interface=CONFIG_BUS_NAME,
                         in_signature='s',
                         out_signature='i')
    def GetInt(self, key):
        section, name = self._split_key(key)
        try:
            return self._confparser.getint(section, name)
        except (configparser.NoSectionError, configparser.NoOptionError):
            raise ValueNotExistError(key)

    @dbus.service.method(dbus_interface=CONFIG_BUS_NAME,
                         in_signature='s',
                         out_signature='d')
    def GetDouble(self, key):
        section, name = self._split_key(key)
        try:
            return self._confparser.getfloat(section, name)
        except (configparser.NoSectionError, configparser.NoOptionError):
            raise ValueNotExistError(key)

    @dbus.service.method(dbus_interface=CONFIG_BUS_NAME,
                         in_signature='s',
                         out_signature='s')
    def GetString(self, key):
        section, name = self._split_key(key)
        try:
            return self._confparser.get(section, name)
        except (configparser.NoSectionError, configparser.NoOptionError):
            raise ValueNotExistError(key)

    @dbus.service.method(dbus_interface=CONFIG_BUS_NAME,
                         in_signature='s',
                         out_signature='as')
    def GetStringList(self, key):
        value = self.GetString(key)
        try:
            return split(value)
        except (configparser.NoSectionError, configparser.NoOptionError):
            raise ValueNotExistError(key)

    def _set_value(self, key, value, overwrite=True):
        section, name = self._split_key(key, True)
        if overwrite or not self._confparser.has_option(section, name):
            self._confparser.set(section, name, str(value))
            self._changed_signals[key] = True
            self._schedule_save()
            self._schedule_signal()

    @dbus.service.method(dbus_interface=CONFIG_BUS_NAME,
                         in_signature='sb',
                         out_signature='')
    def SetBool(self, key, value):
        self._set_value(key, 'true' if value else 'false')

    @dbus.service.method(dbus_interface=CONFIG_BUS_NAME,
                         in_signature='si',
                         out_signature='')
    def SetInt(self, key, value):
        self._set_value(key, value)

    @dbus.service.method(dbus_interface=CONFIG_BUS_NAME,
                         in_signature='sd',
                         out_signature='')
    def SetDouble(self, key, value):
        self._set_value(key, value)

    @dbus.service.method(dbus_interface=CONFIG_BUS_NAME,
                         in_signature='ss',
                         out_signature='')
    def SetString(self, key, value):
        self._set_value(key, value)

    @dbus.service.method(dbus_interface=CONFIG_BUS_NAME,
                         in_signature='sas',
                         out_signature='')
    def SetStringList(self, key, value):
        self._set_value(key, join(value))

    @dbus.service.method(dbus_interface=CONFIG_BUS_NAME,
                         in_signature='a{sv}',
                         out_signature='')
    def SetDefaultValues(self, values):
        for k, v in values.items():
            if isinstance(v, list):
                v = join(v)
            self._set_value(k, v, False)

    def _schedule_save(self, filename=None):
        if self._save_timer is None:
            self._save_timer = GLib.timeout_add(1000,
                                                lambda: self.save(filename))

    def save(self, filename=None):
        if filename is None:
            filename = self._filename
        with open(filename, 'w') as f:
            self._confparser.write(f)
        if self._save_timer is not None:
            GLib.source_remove(self._save_timer)
            self._save_timer = None

    def _schedule_signal(self):
        if self._signal_timer is None:
            self._signal_timer = GLib.timeout_add(500,
                                                  lambda: self.emit_change())

    def emit_change(self):
        if self._signal_timer is not None:
            GLib.source_remove(self._signal_timer)
            self._signal_timer = None
        changed = list(self._changed_signals.keys())
        self.ValueChanged(changed)
        self._changed_signals = {}

    @dbus.service.signal(dbus_interface=CONFIG_BUS_NAME,
                         signature='as')
    def ValueChanged(self, changed):
        pass


def split(value, sep=';'):
    r"""
    >>> split('')
    []
    >>> split(' ')
    [' ']
    >>> split('single')
    ['single']
    >>> split('one;two')
    ['one', 'two']
    >>> split('one;')
    ['one']
    >>> split(';one;two;')
    ['', 'one', 'two']
    >>> split(r'one\;two;three\\;four')
    ['one;two', 'three\\', 'four']
    >>> split(r'\one\\\;two;\\three\\\\;four;')
    ['\\one\\;two', '\\three\\\\', 'four']
    >>> split('; ')
    ['', ' ']
    """
    start = 0
    ret = []
    item = []
    curr = 0
    while curr <= len(value):
        if curr == len(value) or value[curr] == sep:
            if start < curr:
                item.append(value[start:curr])
            if curr != len(value) or item:
                ret.append(''.join(item))
            item = []
            start = curr + 1
        elif value[curr] == '\\' and curr < len(value) - 1:
            tag = value[curr + 1]
            if tag == '\\' or tag == sep:
                item.append(value[start:curr])
                start = curr + 1
                curr = start
        curr = curr + 1
    return ret


def join(values, sep=';'):
    r"""
    >>> join([])
    ''
    >>> join([''])
    ';'
    >>> join(['one'])
    'one;'
    >>> join(['one', 'two'])
    'one;two;'
    >>> join(['one;', 'two'])
    'one\\;;two;'
    >>> print join([r'on\e', 't;wo'])
    on\\e;t\;wo;
    """
    if not values:
        return ''
    result = []
    for item in values:
        result.append(item.replace('\\', '\\\\').replace(sep, '\\;'))
    return sep.join(result) + sep


def test():
    import doctest
    doctest.testmod()


def run():
    app = App('Config')
    if len(sys.argv) > 1:
        ini_conf = IniConfig(app.connection, sys.argv[1])
    else:
        ini_conf = IniConfig(app.connection)  # noqa: F841
    app.run()


if __name__ == '__main__':
    import sys
    if '--test' in sys.argv:
        test()
    else:
        run()
